"""cyme.utils"""
from __future__ import absolute_import

import sys

from importlib import import_module

from celery import current_app as celery
from celery.utils import get_cls_by_name
from kombu.utils.functional import promise, maybe_promise # noqa
from kombu.utils import uuid, cached_property             # noqa
from unipath import Path as _Path

_pkg_cache = {}


def force_list(obj):
    """Force object to be a list.

    If ``obj`` is a scalar value then a list with that value as
    sole element is returned, or
    if ``obj`` is a tuple then it is coerced into a list.

    """
    if isinstance(obj, tuple):
        return list(obj)
    elif not isinstance(obj, list):
        return [obj]
    return obj


def find_package(mod, _s=None):
    """Find the package a module belongs to.

    if you have structure::

        package/__init__.py
               /foo.py
               /bar/__init__.py
               /bar/baz.py

    Then the following examples returns::

        >>> find_package(import_module("package"))
        "package"
        >>> find_package(import_module("package.foo"))
        "package"
        >>> find_package(import_module("package.bar.baz"))
        >>> package.bar

    Note that this does not look at the file system,
    but rather uses the ``__package__`` attribute of a module.

    """
    pkg = None
    _s = _s or mod
    if not mod:
        return
    if mod in _pkg_cache:
        pkg = _pkg_cache[_s] = _pkg_cache[mod]
    else:
        _mod = sys.modules[mod]
        if _mod.__package__:
            pkg = _pkg_cache[_s] = _mod.__package__
    return pkg or find_package('.'.join(mod.split('.')[:-1]), _s)


def find_symbol(origin, sym):
    """Find symbol in module relative by ``origin``.

    E.g. if ``origin`` is an object in the module ``package.foo``,
    then::

        >>> find_symbol(origin, ".bar.my_symbol")

    will return the object ``my_symbol`` from module ``package.bar``.

    """
    return get_cls_by_name(sym,
                package=find_package(getattr(origin, "__module__", None
                                        or origin.__class__.__module__)))


def instantiate(origin, sym, *args, **kwargs):
    """Like :func:`find_symbol` but instantiates the class found
    using ``*args`` and ``**kwargs``."""
    return find_symbol(origin, sym)(*args, **kwargs)


class Path(_Path):
    """:class:`unipath.Path` version that can use the ``/`` operator
    to combine paths::

        >>> p = Path("foo")
        >>> p / "bar" / "baz"
        Path("foo/bar/baz")
    """

    def __div__(self, other):
        return Path(self, other)


def imerge_settings(a, b):
    """Merge two django settings modules,
    keys in ``b`` have precedence."""
    orig = import_module(a.SETTINGS_MODULE)
    for key, value in vars(b).iteritems():
        if not hasattr(orig, key):
            setattr(a, key, value)


def setup_logging(loglevel="INFO", logfile=None):
    """Setup logging using ``loglevel`` and ``logfile``.

    stderr will be used if not logfile provided.

    """
    from celery.utils import LOG_LEVELS
    if isinstance(loglevel, basestring):
        loglevel = LOG_LEVELS[loglevel]
    return celery.log.setup_logging_subsystem(loglevel, logfile)


def redirect_stdouts_to_logger(loglevel="INFO", logfile=None,
        redirect_level="WARNING", stdout=False, stderr=True):
    """See :meth:`celery.log.Log.redirect_stdouts_to_logger`."""
    # XXX Currently unused.
    log = celery.log
    handled = setup_logging(loglevel, logfile)
    if not handled:
        return log.redirect_stdouts_to_logger(
                    log.get_default_logger(name="cyme"),
                    redirect_level, stdout=stdout, stderr=stderr)


class LazyProgressBar(object):

    def __init__(self, size, description=None, endtext=None):
        self.size = size
        self.current = 0
        self.description = description
        self.endtext = endtext
        self._finished = False

    def step(self, i=1, **kwargs):
        if not self._finished:
            if not self.current:
                # implicit start
                self.current = 1
                if self.description:
                    sys.stderr.write("\n\n%s\n" % (self.description, ))
                return self._bar   ## noqa
            self.current += i
            if self.current >= self.size:
                self.size = self._bar.maxval = self.current
            self._bar.update(self.current)

    def finish(self, **kwargs):
        if not self._finished:
            self._bar.finish()
            self._finished = True
            if self.endtext:
                sys.stderr.write(self.endtext)

    @cached_property
    def _bar(self):
        from progressbar import ProgressBar
        return ProgressBar(maxval=self.size).start()



"""cyme.bin.cyme_branch

- This is the script run by the :program:`cyme-branch` script installed
  by the Cyme distribution (defined in setup.py's ``entry_points``).

- It in turn executes the cyme-branch management command.

"""

from __future__ import absolute_import

from .base import app


@app(needs_eventlet=True)
def cyme_branch(env, argv):
    from ..management.commands import cyme_branch
    cyme_branch.Command(env).run_from_argv([argv[0], "cyme-branch"] + argv[1:])


if __name__ == "__main__":
    cyme_branch()



import os
import sys

DB_NAME = os.environ["CYME_DB_NAME"] = "funtest.db"

sys.path.insert(0, os.path.join(os.getcwd(), os.pardir))


def teardown():
    if os.path.exists(DB_NAME):
        os.unlink(DB_NAME)



"""cyme.branch.thread

- Utilities for working with greenlets.

"""

from __future__ import absolute_import
from __future__ import with_statement

import os

from Queue import Empty

from cl.g import Event, spawn, timer, Queue
from kombu.log import LogMixin
from eventlet import Timeout

from . import signals


class AlreadyStartedError(Exception):
    """Raised if trying to start a thread instance that is already started."""


class gThread(LogMixin):
    AlreadyStarted = AlreadyStartedError
    Timeout = Timeout

    #: Name of the thread, used in logs and such.
    name = None

    #: Greenlet instance of the thread, set when the thread is started.
    g = None

    #: Set when the thread is requested to stop.
    should_stop = False

    #: Set this to False if it is not possible to join the thread.
    joinable = True

    _exit_event = None
    _ping_queue = None
    _timers = None
    extra_startup_steps = 0
    extra_shutdown_steps = 0

    def __init__(self):
        self.name = self.name or self.__class__.__name__
        # the exit event is sent when the thread exits,
        # and used by `join` to detect this.
        self._exit_event = Event()

        # we maintain a list of timers started by the thread,
        # so these can be cancelled at shutdown.
        self._timers = []

    def before(self):
        """Called at the beginning of :meth:`start`."""
        pass

    def run(self):
        raise NotImplementedError("gThreads must implement 'run'")

    def after(self):
        """Call after the thread has shut down."""
        pass

    def start_periodic_timer(self, interval, fun, *args, **kwargs):
        """Apply function every ``interval`` seconds.

        :param interval: Interval in seconds (int/float).
        :param fun: The function to apply.
        :param \*args: Additional arguments to pass.
        :keyword \*\*kwargs: Additional keyword arguments to pass.

        :returns: entry object, with ``cancel`` and ``kill`` methods.

        """
        entry = timer(interval, fun, *args, **kwargs)
        self._timers.append(entry)
        return entry

    def start(self):
        """Spawn green thread, and returns
        :class:`~eventlet.greenthread.GreenThread` instance."""
        if self.g is None:
            self.before()
            signals.thread_pre_start.send(sender=self)
            self._ping_queue = Queue()
            g = self.g = self.spawn(self._crashsafe, self.run)
            g.link(self._on_exit)
            self.debug("%s spawned", self.name)
            signals.thread_post_start.send(sender=self)
            return g
        raise self.AlreadyStarted("can't start thread twice")

    def stop(self, join=True, timeout=1e100):
        """Shutdown the thread.

        This will also cancel+kill any periodic timers registered
        by the thread.

        :keyword join: Given that the thread is :attr:`joinable`, if
            true will also wait until the thread exits
            (by calling :meth:`join`).
        :keyword timeout: Timeout for join (default is 1e+100).

        """
        self.debug("shutdown initiated")
        signals.thread_pre_shutdown.send(sender=self)
        self.should_stop = True
        for entry in self._timers:
            self.debug("killing timer %r" % (entry, ))
            entry.cancel()
            entry.kill()
        if join and self.joinable:
            try:
                self.join(timeout)
            except self.Timeout:
                self.error(
                    "exceeded exit timeout (%s), will try to kill", timeout)
                self.kill()
        self.after()
        signals.thread_post_shutdown.send(sender=self)

    def join(self, timeout=None):
        """Wait until the thread exits.

        :keyword timeout: Timeout in seconds (int/float).

        :raises eventlet.Timeout: if the thread can't be joined
            before the provided timeout.


        """
        with self.Timeout(timeout):
            signals.thread_pre_join.send(sender=self, timeout=timeout)
            self.debug("joining (%s)", timeout)
            self._exit_event.wait()
            signals.thread_post_join.send(sender=self)

    def respond_to_ping(self):
        try:
            self._ping_queue.get_nowait().send()
        except Empty:
            pass

    def _do_ping(self, timeout=None):
        with self.Timeout(timeout):
            event = Event()
            self._ping_queue.put_nowait(event)
            event.wait()
            return event.ready()

    def ping(self, timeout=None):
        if self.g and self._ping_queue is not None:
            return self._do_ping(timeout)
        return False

    def _on_exit(self, g):
        # called when the thread exits to unblock `join`.
        self._exit_event.send()

    def kill(self):
        """Kill the green thread."""
        if self.g is not None:
            self.g.kill()

    def _crashsafe(self, fun, *args, **kwargs):
        try:
            fun(*args, **kwargs)
            signals.thread_exit.send(sender=self)
            self.debug("exiting")
        except Exception, exc:
            self.error("Thread crash detected: %r", exc)
            os._exit(0)
        except self.Timeout, exc:
            self.error("Thread raised timeout: %r", exc)
            os._exit(0)

    def spawn(self, fun, *args, **kwargs):
        return spawn(fun, *args, **kwargs)

    @property
    def logger_name(self):
        return self.name



"""cyme.branch.state

- Global branch state.

- Used to keep track lost connections and so on, which is used by the
  supervisor to know if an instance is actually down, or if it is just the
  connection being shaky.

"""

from __future__ import absolute_import

from time import time

from ..utils import cached_property, find_symbol


class State(object):
    broker_last_revived = None

    #: set to true if the process is a cyme-branch
    is_branch = False

    def on_broker_revive(self, *args, **kwargs):
        self.broker_last_revived = time()
        self.supervisor.resume()

    @property
    def time_since_broker_revived(self):
        return time() - self.broker_last_revived

    @cached_property
    def supervisor(self):
        return find_symbol(self, ".supervisor.supervisor")

state = State()



# -*- coding: utf-8 -*-

import sys
import os

this = os.path.dirname(os.path.abspath(__file__))

# If your extensions are in another directory, add it here. If the directory
# is relative to the documentation root, use os.path.abspath to make it
# absolute, like shown here.
sys.path.append(os.path.join(os.pardir, "tests"))
sys.path.append(os.path.join(this, "_ext"))
from django.core.management import setup_environ
import settings
setup_environ(settings)
import cyme


# use app loader
from celery import Celery
app = Celery(set_as_current=True)
app.conf.update(BROKER_TRANSPORT="memory",
                CELERY_RESULT_BACKEND="cache",
                CELERY_CACHE_BACKEND="memory",
                CELERYD_HIJACK_ROOT_LOGGER=False,
                CELERYD_LOG_COLOR=False)

# General configuration
# ---------------------

extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.coverage',
              'sphinx.ext.pngmath',
              'sphinx.ext.intersphinx',
              'sphinxcontrib.issuetracker',
              'celerydocs']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['.templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'Cyme'
copyright = u'2011, VMware, Inc.'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = ".".join(map(str, cyme.VERSION[0:2]))
# The full version, including alpha/beta/rc tags.
release = cyme.__version__

exclude_trees = ['.build']

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'trac'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['.static']

html_use_smartypants = True

# If false, no module index is generated.
html_use_modindex = True

# If false, no index is generated.
html_use_index = True

latex_documents = [
  ('index', 'cyme.tex', ur'Cyme Documentation',
   ur'VMWare, Inc.', 'manual'),
]

html_show_sphinx = False
intersphinx_mapping = {
        "http://docs.python.org/dev": None,
        "http://kombu.readthedocs.org/en/latest/": None,
        "http://django-celery.readthedocs.org/en/latest": None,
}

html_theme = "celery"
html_theme_path = ["_theme"]
html_sidebars = {
    'index': ['sidebarintro.html', 'sourcelink.html', 'searchbox.html'],
    '**': ['sidebarlogo.html', 'relations.html',
           'sourcelink.html', 'searchbox.html'],
}

### Issuetracker

#issuetracker = "github"
#issuetracker_user = "ask"
#issuetracker_project = "celery"
#issuetracker_issue_pattern = r'[Ii]ssue #(\d+)'



"""cyme.bin.cyme_list_branches"""
from __future__ import absolute_import

from .base import app

import sys
import anyjson


def get_branches(broker=None, limit=None):
    from cyme.branch.controller import Branch
    from celery import current_app as celery
    if limit:
        limit = int(limit)

    args = [broker] if broker else []
    conn = celery.broker_connection(*args)
    return Branch(connection=conn).all(limit=limit)


@app()
def main(env, argv):
    opts = dict(arg.split("=", 1)
                    for arg in argv[1:] if arg.startswith("--"))
    print(anyjson.serialize(get_branches(opts.get("--broker"),
                                         opts.get("--limit"))))


if __name__ == "__main__":
    main()





"""cyme.conf"""

from __future__ import absolute_import

from django.conf import settings

from .utils import Path

CYME_INSTANCE_DIR = Path(getattr(settings,
                        "CYME_INSTANCE_DIR", "instances")).absolute()
CYME_DEFAULT_POOL = getattr(settings, "CYME_DEFAULT_POOL", "processes")



import os
import sys

# funtest config
sys.path.insert(0, os.getcwd())
sys.path.insert(0, os.path.join(os.getcwd(), os.pardir))

from cyme.bin.base import app, Env
from cyme.utils import cached_property, Path, uuid
from eventlet.event import Event
from nose import SkipTest

from .utils import unittest

CYME_PORT = int(os.environ.get("CYME_PORT") or 8013)
CYME_URL = "http://127.0.0.1:%s" % (CYME_PORT, )
CYME_INSTANCE_DIR = Path("instances").absolute()

_branch = [None]


@app(needs_eventlet=True, instance_dir=CYME_INSTANCE_DIR)
def start_branch(env, argv=None):
    env.syncdb(interactive=False)
    from cyme.branch import Branch
    ready_event = Event()
    CYME_INSTANCE_DIR.mkdir()
    instance = Branch("127.0.0.1:%s" % (CYME_PORT, ), numc=1,
                      ready_event=ready_event)
    instance.start()
    ready_event.wait()
    return instance


def destroy_branch(branch):
    branch.stop()


def teardown():
    if _branch[0] is not None:
        destroy_branch(_branch[0])
    if CYME_INSTANCE_DIR.isdir():
        CYME_INSTANCE_DIR.rmtree()


class ClientTestCase(unittest.TestCase):

    @cached_property
    def Client(self):
        from cyme import Client
        return Client


class BranchTestCase(unittest.TestCase):

    def setUp(self):
        if _branch[0] is None:
            _branch[0] = start_branch()


class test_create_app(BranchTestCase, ClientTestCase):

    def test_create(self):
        client = self.Client(CYME_URL)
        app = client.add(uuid())
        self.assertTrue(repr(app))
        self.assertTrue(app)
        self.assertTrue(app.info)
        self.assertTrue(app.info.broker)
        self.assertIn(app.app, app.all())
        app = client.get(app.app)
        self.assertTrue(app)
        app.delete()
        self.assertNotIn(app.app, app.all())


class test_basic(BranchTestCase, ClientTestCase):

    def setUp(self):
        BranchTestCase.setUp(self)
        self.app = self.Client(CYME_URL).add(uuid())

    def tearDown(self):
        self.app.delete()
        BranchTestCase.tearDown(self)

    def test_basic(self):
        app = self.app
        instance = app.instances.add()
        self.assertTrue(repr(instance))
        self.assertTrue(instance)
        instance = app.instances.get(instance.name)
        self.assertTrue(instance)
        self.assertIn(instance, app.instances)

        q = uuid()
        expected = dict(exchange=q, exchange_type="topic", routing_key=q)
        queue = app.queues.add(q, **expected)
        self.assertTrue(repr(queue))
        queue = app.queues.get(queue.name)
        self.assertTrue(queue)
        for key, value in expected.items():
            self.assertEqual(getattr(queue, key), value)
        self.assertIn(queue, app.queues)

        self.assertTrue(instance.consumers.add(queue))
        self.assertTrue(instance.consumers.delete(queue))

        queue.delete()
        self.assertNotIn(queue, app.queues)

        instance.delete()
        self.assertNotIn(instance, app.instances)


if __name__ == "__main__":
    unittest.main()



"""cyme.admin

- Django Admin interface extensions for Cyme.

"""
from __future__ import absolute_import

from django.contrib import admin
from django.utils.html import escape
from django.utils.translation import ugettext_lazy as _

from djcelery.admin_utils import action, display_field, fixedwidth
from djcelery.utils import naturaldate

from .models import Broker, Instance, Queue
from .branch.supervisor import supervisor


@display_field(_("max/min concurrency"), "max_concurrency")
def maxmin_concurrency(instance):
    return "%s / %s" % (instance.max_concurrency, instance.min_concurrency)


@display_field(_("created"), "created_at")
def created_at(instance):
    return """<div title="%s">%s</div>""" % (
                escape(str(instance.created_at)),
                escape(naturaldate(instance.created_at)))


@display_field(_("status"), "is_enabled")
def status(instance):
    enabled = "Enabled" if instance.is_enabled else "Disabled"
    if instance.alive():
        state, color = "ONLINE", "green"
    else:
        state, color = "OFFLINE", "red"
    return """<b>%s (<span style="color: %s;">%s</span>)</b>""" % (
            enabled, color, state)


class InstanceAdmin(admin.ModelAdmin):
    detail_title = _("Instance detail")
    list_page_title = _("Instances")
    date_hierarchy = "created_at"
    fieldsets = (
            (None, {
                "fields": ("name", "max_concurrency", "min_concurrency",
                           "_queues", "is_enabled", "_broker"),
                "classes": ("extrapretty", ),
            }), )
    list_display = (fixedwidth("name", pt=10), maxmin_concurrency,
                    status, "broker")
    read_only_fields = ("created_at", )
    list_filter = ("name", "max_concurrency", "min_concurrency", "_queues")
    search_fields = ("name", "max_concurrency", "min_concurrency", "_queues")
    actions = ["disable_instances",
               "enable_instances",
               "restart_instances"]

    @action(_("Disable selected instances"))
    def disable_instances(self, request, queryset):
        for instance in queryset:
            instance.disable()
        supervisor.verify(queryset).wait()

    @action(_("Enable selected instances"))
    def enable_instances(self, request, queryset):
        for instance in queryset:
            instance.enable()
        supervisor.verify(queryset).wait()

    @action(_("Restart selected instances"))
    def restart_instances(self, request, queryset):
        supervisor.restart(queryset).wait()


admin.site.register(Broker)
admin.site.register(Instance, InstanceAdmin)
admin.site.register(Queue)



"""cyme.client

- Python client for the Cyme HTTP API.

Branches
~~~~~~~~

   >>> client = Client("http://localhost:8000")
   >>> client.branches
   ["cyme1.example.com", "cyme2.example.com"]

   >>> client.branch_info("cyme1.example.com")
   {'sup_interval': 5, 'numc': 2, 'loglevel': 'INFO',
    'logfile': None, 'id': 'cyme1.example.com', 'port': 8000}

Applications
~~~~~~~~~~~~

    >>> app = client.get("foo")
    >>> app
    <Client: 'http://localhost:8016/foo'>
    >>> app.info
    {'name': 'foo', 'broker': 'amqp://guest:guest@localhost:5672//'}

Queues
~~~~~~

    >>> app.queues.add("myqueue", exchange="myex", routing_key="x")
    >>> <Queue: u'myqueue'>
    >>> my_queue = app.queues.get("my_queue")

    >>> app.queues
    [<Queue: u'myqueue'>]


Instances
~~~~~~~~~

    >>> i = app.instances.add()
    >>> i
    <Instance: u'd87798f3-0bb0-4161-8e0b-a5f069b1d58b'>
    >>> i.name
    u'd87798f3-0bb0-4161-8e0b-a5f069b1d58b'

    >>> i.broker  # < inherited from app
    u'amqp://guest:guest@localhost:5672//'

    >>> app.instances
    [<Instance: u'd87798f3-0bb0-4161-8e0b-a5f069b1d58b'>]

    >>> i.autoscale()   # current autoscale settings
    {'max': 1, 'min': 1}

    >>> i.autoscale(max=10, min=10)  # always run 10 processes
    {'max': 10, 'min': 10}

    >>> i.stats()
    {'total': {},
     'consumer': {'prefetch_count': 80,
                  'broker': {'transport_options': {},
                             'login_method': 'AMQPLAIN',
                             'hostname': '127.0.0.1',
                             'userid': 'guest',
                             'insist': False,
                             'connect_timeout': 4,
                             'ssl': False,
                             'virtual_host': '/',
                             'port': 5672,
                             'transport': 'amqp'}},
    'pool': {'timeouts': [None, None],
             'processes': [76003],
             'max-concurrency': 1,
             'max-tasks-per-child': None,
             'put-guarded-by-semaphore': True},
    'autoscaler': {'current': 1, 'max': 1, 'min': 1, 'qty': 0}}

Consumers
~~~~~~~~~

    >>> instance.consumers.add(my_queue)
    {"ok": "ok"}

    >>> instance_consumers.delete(my_queue)
    {"ok": "ok"}

    >>> instance.consumers
    #... consumers with full declarations ...


Deleting
~~~~~~~~

This will delete the queue and eventually force all worker instances
to stop consuming from it::

    >>> my_queue.delete()


This will shutdown and delete an instance::

    >>> i.delete()


"""

from __future__ import absolute_import

from dictshield import fields

from . import base
from .base import Path
from ..utils import cached_property

# XXX `requests` does not currently seem to support using the
#     data argument with PUT requests.


class Instance(base.Model):
    name = fields.StringField(max_length=200)
    broker = fields.StringField(max_length=200)
    pool = fields.StringField(max_length=200)
    min_concurrency = fields.IntField()
    max_concurrency = fields.IntField()
    is_enabled = fields.BooleanField()
    queue_names = fields.ListField(fields.StringField(max_length=200))
    arguments = fields.StringField(max_length=200)
    extra_config = fields.StringField(max_length=200)

    def __repr__(self):
        return "<Instance: %r>" % (self.name, )


class Queue(base.Model):
    name = fields.StringField(max_length=200, required=True)
    exchange = fields.StringField(max_length=200, required=False)
    exchange_type = fields.StringField(max_length=200, required=False)
    routing_key = fields.StringField(max_length=200, required=False)
    options = fields.StringField(max_length=None)

    def __repr__(self):
        return "<Queue: %r>" % (self.name, )


class Client(base.Client):
    app = "cyme"

    class Instances(base.Section):

        class Model(Instance):

            class Consumers(base.Section):

                def __init__(self, client, name):
                    base.Section.__init__(self, client)
                    self.path = self.client.path / name / "queues"

                def create_model(self, data, *args, **kwargs):
                    return data

            def __init__(self, *args, **kwargs):
                base.Model.__init__(self, *args, **kwargs)
                self.consumers = self.Consumers(self.parent, name=self.name)
                self.path = self.parent.path / self.name

            def stats(self):
                return self.parent.stats(self.name)

            def autoscale(self, max=None, min=None):
                return self.parent.autoscale(self.name, max=max, min=min)

            class LazyQueues(object):

                def __init__(self, instance):
                    self.instance = instance

                def __iter__(self):
                    return self.instance._get_queues()

                def __getitem__(self, index):
                    return self._eval[index]

                def __contains__(self, queue):
                    if isinstance(queue, Queue):
                        queue = queue.name
                    return queue in self.instance.queue_names

                def __len__(self):
                    return len(self._eval)

                def __repr__(self):
                    return repr(self._eval)

                @cached_property
                def _eval(self):
                    return list(iter(self))

            def _get_queues(self):
                return (self.parent.client.queues.get(name)
                            for name in self.queue_names)

            @property
            def queues(self):
                return self.LazyQueues(self)

        def add(self, name=None, broker=None, arguments=None,
                config=None, nowait=False):
            # name is optional for instances
            return base.Section.add(self, name, nowait,
                                    broker=broker,
                                    arguments=arguments,
                                    extra_config=config)

        def stats(self, name):
            return self.GET(self.path / name / "stats")

        def autoscale(self, name, max=None, min=None):
            return self.POST(self.path / name / "autoscale",
                             params={"max": max, "min": min})

        def create_model(self, data, *args, **kwargs):
            data["queue_names"] = data.pop("queues", None)
            return base.Section.create_model(self, data, *args, **kwargs)

    class Queues(base.Section):

        class Model(Queue):
            pass

        def add(self, name, exchange=None, exchange_type=None,
                routing_key=None, nowait=False, **options):
            options = self.serialize(options) if options else None
            return base.Section.add(self, name, nowait,
                                    exchange=exchange,
                                    exchange_type=exchange_type,
                                    routing_key=routing_key,
                                    options=options)

    def __init__(self, url=None, app=None, info=None):
        super(Client, self).__init__(url)
        self.app = app
        self.instances = self.Instances(self)
        self.queues = self.Queues(self)
        self.info = info or {}

    def add(self, name, broker=None, arguments=None, extra_config=None,
            nowait=False):
        return self.create_model(name, self.root("POST",
                                 self.maybe_async(name, nowait),
                                 data={"broker": broker,
                                       "arguments": arguments,
                                       "extra_config": extra_config}))

    def get(self, name=None):
        return self.create_model(name, self.root("GET", name or self.app))

    def delete(self, name=None):
        return self.root("DELETE", name or self.app)

    @property
    def branches(self):
        return self.root("GET", "branches")

    def branch_info(self, id):
        return self.root("GET", Path("branches") / id)

    def all(self):
        return self.root("GET")

    def build_url(self, path):
        url = self.url
        if self.app:
            url += "/" + self.app
        return url + str(path)

    def create_model(self, name, info):
        return self.clone(app=name, info=base.AttributeDict(info))

    def clone(self, app=None, info=None):
        return self.__class__(url=self.url, app=app, info=info)

    def __repr__(self):
        return "<Client: %r>" % (self.build_url(""), )



"""

.. program:: cyme

``cyme``
========

Cyme management utility

Options
-------

.. cmdoption:: -a, --app

    Application to use.  Required for all operations except for
    when creating, deleting or listing apps.

.. cmdoption:: -n, --nowait

    Don't want for operations to complete (async).

.. cmdoption:: -F, --format

    Output format: pretty (default) or json.

.. cmdoption:: -L, --local

    Flag that if set means that operations will be performed
    locally for the current branch only.
    This means the instance directory must be properly set.

.. cmdoption:: -l, --loglevel

    Set custom log level. One of DEBUG/INFO/WARNING/ERROR/CRITICAL.
    Default is INFO.

.. cmdoption:: -f, --logfile

    Set custom logfile path. Default is :file:`<stderr>`

.. cmdoption:: -D, --instance-dir

    Custom instance directory (default is :file:`instances/`)
    Must be readable by the current user.

    This needs to be properly specified if the :option:`-L` option
    is used.

.. cmdoption:: -b, --broker

    Broker to use for a local `branches` request.

"""

from __future__ import absolute_import

import anyjson
import os
import pprint

from functools import partial
from inspect import getargspec


from celery import current_app as celery
from cyme.client import Client
from cyme.client.base import Model
from cyme.utils import cached_property, instantiate

from .base import CymeCommand, Option, die


try:
    import json as _json   # Python 2.6+
    json_pretty = partial(_json.dumps, indent=4)
except ImportError:
    json_pretty = anyjson.serialize  # noqa


class I(object):

    def __init__(self, app=None, format=None, nowait=False,
            url=None, **kwargs):
        self.url = url
        self.app = app
        self.format = format or "pretty"
        self.nowait = nowait
        self.actions = {
            "branches": {
                "all": self.all_branches,
            },
            "apps": {
                "all": self.all_apps,
                "get": self.get_app,
                "add": self.add_app,
                "delete": self.delete_app},
            "instances": {
                "all": self.all_instances,
                "get": self.get_instance,
                "add": self.add_instance,
                "delete": self.delete_instance,
                "stats": self.instance_stats,
                "autoscale": self.instance_autoscale},
            "queues": {
                "all": self.all_queues,
                "get": self.get_queue,
                "add": self.add_queue,
                "delete": self.delete_queue},
            "consumers": {
                "all": self.all_consumers,
                "add": self.add_consumer,
                "delete": self.delete_consumer},
        }
        self.needs_app = ("instances", "queues")
        self.formats = {"jsonp": json_pretty,
                        "json": anyjson.serialize,
                        "pprint": pprint.pformat}

    def getsig(self, fun, opt_args=None):
        spec = getargspec(fun)
        args = spec.args[:-len(spec.defaults) if spec.defaults else None]
        if args[0] == "self":
            args = args[1:]
        if spec.defaults:
            opt_args = dict(zip(spec.args[len(spec.defaults):], spec.defaults))
        return len(args), args, opt_args

    def _ni(self, *args, **kwargs):
        raise NotImplementedError("subclass responsibility")
    all_apps = get_app = add_app = delete_app = \
        all_instances = get_instances = add_instance = delete_instance = \
            instance_stats = instance_autoscale = \
                all_queues = get_queue = add_queue = delete_queue = \
                    all_consumers = add_consumer = delete_consumer = _ni

    def DISPATCH(self, fqdn, *args):
        type, _, action = fqdn.partition('.')
        if not self.app and type in self.needs_app:
            die("Need to specify --app")
        try:
            handler = self.actions[type][action or "all"]
        except KeyError:
            if type:
                die("No action %r for type %r" % (action, type))
            die("Missing type")
        try:
            response = handler(*args)
        except TypeError:
            raise
            arity, args, optargs = self.getsig(handler)
            die("%s.%s requires %s argument%s: %s %s" % (
                type, action, arity, "s" if arity > 1 else "",
                " ".join(args), self.format_optargs(optargs)))
        return self.format_response(self.prepare_response(response))

    def format_optargs(self, optargs):
        if optargs:
            return " ".join("[%s]" % (k, ) for k in optargs.keys())
        return ''

    def format_response(self, ret):
        return self.formats[self.format](ret)

    def prepare_response(self, ret):
        return ret


class WebI(I):

    def all_branches(self):
        return list(self.client.branches)

    def all_apps(self):
        return list(self.client.all())

    def get_app(self, name):
        return self.client.get(name).info

    def add_app(self, name, broker=None, arguments=None, extra_config=None):
        return self.client.add(name, nowait=self.nowait,
                               broker=broker, arguments=arguments,
                               extra_config=extra_config).info

    def delete_app(self, name):
        return self.client.delete(name, nowait=self.nowait)

    def all_instances(self):
        return list(self.client.instances.all())

    def get_instance(self, name):
        return self.client.instances.get(name)

    def add_instance(self, name=None, broker=None, arguments=None,
            extra_config=None):
        return self.client.instances.add(name=name, broker=broker,
                                         arguments=arguments,
                                         extra_config=extra_config,
                                         nowait=self.nowait)

    def delete_instance(self, name):
        return self.client.instances.get(name).delete(nowait=self.nowait)

    def instance_stats(self, name):
        return self.client.instances.get(name).stats()

    def instance_autoscale(self, name, max=None, min=None):
        return self.client.instances.get(name).autoscale(max=max, min=min)

    def all_consumers(self, instance_name):
        return list(self.client.instances.get(instance_name).consumers)

    def add_consumer(self, instance_name, queue_name):
        return self.client.instances.get(instance_name)\
                    .consumers.add(queue_name, nowait=self.nowait)

    def delete_consumer(self, instance_name, queue_name):
        return self.client.instances.get(instance_name)\
                    .consumers.delete(queue_name, nowait=self.nowait)

    def all_queues(self):
        return list(self.client.queues.all())

    def get_queue(self, name):
        return self.client.queues.get(name)

    def add_queue(self, name, exchange=None, exchange_type=None,
            routing_key=None, options=None):
        options = anyjson.deserialize(options) if options else {}
        return self.client.queues.add(name, exchange=exchange,
                                            exchange_type=exchange_type,
                                            routing_key=routing_key,
                                            nowait=self.nowait,
                                            **options)

    def delete_queue(self, name):
        return self.client.queues.delete(name, nowait=self.nowait)

    def _part(self, p):
        if isinstance(p, Model):
            return dict((k, v) for k, v in p.to_python().iteritems()
                            if not k.startswith("_"))
        return p

    def prepare_response(self, ret):
        if isinstance(ret, (list, tuple)):
            return map(self._part, ret)
        return self._part(ret)

    @cached_property
    def client(self):
        client = Client(self.url)
        if self.app:
            return client.get(self.app)
        return client


class LocalI(I):

    def __init__(self, *args, **kwargs):
        super(LocalI, self).__init__(*args, **kwargs)
        self.broker = kwargs.get("broker")
        self.limit = kwargs.get("limit")
        from cyme.branch.controller import apps, instances, queues
        self.get_app = apps.get
        self.apps = apps.state
        self.instances = instances.state
        self.queues = queues.state

    def all_branches(self):
        from cyme.branch.controller import Branch
        args = [self.broker] if self.broker else []
        conn = celery.broker_connection(*args)
        return Branch(connection=conn).all(limit=self.limit)

    def all_apps(self):
        return [app.as_dict() for app in self.apps.objects.all()]

    def get_app(self):
        return self.apps.objects.get(app=self.app)

    def add_app(self, name, broker=None, arguments=None, extra_config=None):
        return self.apps.add(name, broker=broker, arguments=arguments,
                                   extra_config=extra_config)

    def delete_app(self, name):
        self.apps.delete(name)
        return {"ok": "ok"}

    def all_instances(self):
        return [instance.as_dict()
                    for instance in self.instances.objects.filter(
                        app=self.get_app(self.app))]

    def get_instance(self, name):
        return self.instances.get(name, app=self.app)

    def add_instance(self, name=None, broker=None, arguments=None,
            extra_config=None):
        return self.instances.add(name, broker=broker, app=self.app,
                                  arguments=arguments,
                                  extra_config=extra_config)

    def delete_instance(self, name):
        return {"ok": self.instances.remove(name)}

    def instance_stats(self, name):
        return self.instances.stats(name)

    def _get_instance(self, name):
        return self.instances.objects.get(name=name)

    def instance_autoscale(self, name, max=None, min=None):
        return dict(zip(["max", "min"],
                    self._get_instance(name=name)._update_autoscale(max, min)))

    def all_consumers(self, instance_name):
        return self._get_instance(name=instance_name).consuming_from()

    def add_consumer(self, instance_name, queue_name):
        self._get_instance(name=instance_name)\
                .add_queue_eventually(queue_name)
        return {"ok": "ok"}

    def delete_consumer(self, instance_name, queue_name):
        self._get_instance(name=instance_name)\
                .remove_queue_eventually(queue_name)
        return {"ok": "ok"}

    def all_queues(self):
        return [queue.as_dict()
                    for queue in self.queues.objects.all()]

    def get_queue(self, name):
        return self.queues.get(name)

    def add_queue(self, name, exchange=None, exchange_type=None,
            routing_key=None, options=None):
        options = anyjson.deserialize(options) if options else {}
        return self.queues.add(name, exchange=exchange,
                                     exchange_type=exchange_type,
                                     routing_key=routing_key,
                                     **options)

    def delete_queue(self, name):
        self.instances.remove_queue_from_all(name)
        return {"ok": self.queues.delete(name)}


class Command(CymeCommand):
    name = "cyme"
    args = """type command [args]
E.g.:
    cyme apps
    cyme apps.add <name> [broker URL] [arguments] [extra config]
    cyme apps.[get|delete] <name>

    cyme -a <app> instances
    cyme -a <app> instances.add [name] [broker URL] [arguments] [extra config]
    cyme -a <app> instances.[get|delete|stats] <name>
    cyme -a <app> instances.autoscale <name> [max] [min]

    cyme -a <app> queues
    cyme -a <app> queues.add <name> [exchange] [type] [rkey] [opts]
    cyme -a <app> queues.[get|delete] <name>

    cyme start-all
    cyme shutdown-all

    cyme createsuperuser

    cyme shell
    """
    option_list = CymeCommand.option_list + (
       Option('-a', '--app',
              default=None, action="store", dest="app",
              help="application to use"),
       Option("-n", "--nowait",
              default=False, action="store_true", dest="nowait",
              help="Don't want for operations to complete (async)."),
       Option("-F", "--format",
              default="jsonp", action="store", dest="format",
              help="Output format: jsonp (default) json or pprint"),
       Option("-L", "--local",
              default=False, action="store_true", dest="local",
              help="Perform operations locally for this branch only."),
       Option('-l', '--loglevel',
              default="WARNING", action="store", dest="loglevel",
              help="Choose between DEBUG/INFO/WARNING/ERROR/CRITICAL"),
       Option('-D', '--instance-dir',
              default=None, action="store", dest="instance_dir",
              help="Custom instance dir. Default is instances/"),
       Option('-u', '--url',
              default="http://localhost:8000", dest="url",
              help="Custom URL"),
       Option('-b', '--broker',
              default="None", dest="broker",
              help="Broker to use for a local branches request"),
    )

    help = 'Cyme management utility'

    def handle(self, *args, **kwargs):
        local = kwargs.pop("local", False)
        kwargs = self.prepare_options(**kwargs)
        self.commands = {"shell": self.drop_into_shell,
                         "sh": self.drop_into_shell,
                         "start-all": self.start_all,
                         "restart-all": self.restart_all,
                         "shutdown-all": self.shutdown_all,
                         "createsuperuser": self.create_superuser}
        self.setup_logging(**kwargs)
        args = list(args)
        self.enter_instance_dir()
        if local:
            self.env.syncdb()
        if args:
            if args[0] in self.commands:
                return self.commands[args[0]]()
            print((LocalI if local else WebI)(**kwargs).DISPATCH(*args))
        else:
            self.print_help()

    def start_all(self):
        self.status.start_all()

    def restart_all(self):
        self.status.restart_all()

    def shutdown_all(self):
        self.status.shutdown_all()

    def drop_into_shell(self):
        from cyme.utils import setup_logging
        if os.environ.get("CYME_LOG_DEBUG"):
            setup_logging("DEBUG")
        self.env.management.call_command("shell")

    def create_superuser(self):
        self.env.management.call_command("createsuperuser")

    @cached_property
    def status(self):
        return instantiate(self, "cyme.status.Status")



from __future__ import absolute_import

from celery.tests.utils import unittest
from mock import Mock

from cyme.models import Instance, Queue


class test_Queue(unittest.TestCase):

    def test__unicode__(self):
        self.assertTrue(unicode(Queue(name="foo")))

    def test_enabled(self):
        q1 = Queue.objects.add("foo", options={"x": 1})
        q2 = Queue.objects.add("bar")
        q2.is_enabled = False
        q2.save()

        x = Queue.objects.enabled()
        self.assertEqual(x[0], q1)


class test_Instance(unittest.TestCase):

    def setUp(self):
        self._mt, Instance.MultiTool = Instance.MultiTool, Mock()
        self._query, Instance._query = Instance._query, Mock()

    def tearDown(self):
        Instance.MultiTool = self._mt
        Instance._query = self._query
        Queue.objects.all().delete()
        Instance.objects.all().delete()

    def test__unicode__(self):
        self.assertTrue(unicode(Instance(name="foo")))

    def test_add(self):
        n1 = Instance.objects.add()
        self.assertTrue(n1.name)
        self.assertTrue(n1.default_args)
        self.assertTrue(n1.direct_queue)
        self.assertTrue(n1.multi)
        self.assertTrue(n1.argtuple)
        self.assertTrue(n1.argv)
        self.assertIsNone(n1.getpid())

        n2 = Instance.objects.add("foo")
        self.assertEqual(n2.name, "foo")

        n1.disable()
        self.assertFalse(Instance.objects.get(name=n1.name).is_enabled)
        n1.enable()
        self.assertTrue(Instance.objects.get(name=n1.name).is_enabled)

    def test_start_stop_restart(self):
        n = Instance.objects.add()
        n.start()
        n.stop()
        n.restart()

    def test_add_cancel_queue(self):
        n = Instance.objects.add()
        q = Queue.objects.create(name="xiziasd")
        q.save()

        n.add_queue(q)
        n._query.assert_called_with("add_consumer", dict(queue=q.name,
                                    exchange=q.name, routing_key=q.name,
                                    exchange_type=None))

        n.cancel_queue(q)
        n._query.assert_called_with("cancel_consumer", dict(queue=q.name))

    def test_objects(self):
        n = Instance.objects.add()
        Instance.objects.disable(n)
        Instance.objects.enable(n)
        Instance.objects.remove(n)

    def test_with_queues(self):
        Instance.objects.add(queues="foo,bar,baz")
        Instance.objects.add_queue("foo")
        Instance.objects.add_queue("xaz")

        Instance.objects.remove_queue("foo")
        Instance.objects.remove_queue("xaz")



"""cyme.status"""

from __future__ import absolute_import, with_statement

from collections import defaultdict

from celery.datastructures import TokenBucket
from celery.utils.timeutils import rate
from kombu.common import insured as _insured
from kombu.log import LogMixin
from kombu.utils import fxrangemax

from .models import Instance
from .branch.state import state


class Status(LogMixin):
    paused = False
    restart_max_rate = "100/s"

    def __init__(self):
        self._buckets = defaultdict(lambda: TokenBucket(
                                        rate(self.restart_max_rate)))

    def start_all(self):
        for instance in self.all_instances():
            self._do_verify_instance(instance, ratelimit=False)

    def restart_all(self):
        for instance in self.all_instances():
            self._do_restart_instance(instance, ratelimit=False)

    def shutdown_all(self):
        for instance in self.all_instances():
            try:
                self._do_stop_verify_instance(instance)
            except Exception, exc:
                self.error("Couldn't stop instance %s: %r" % (instance.name,
                                                              exc))

    def all_instances(self):
        return Instance.objects.all()

    def insured(self, instance, fun, *args, **kwargs):
        """Ensures any function performing a broadcast command completes
        despite intermittent connection failures."""

        def errback(self, exc, interval):
            self.error(
                "Error while trying to broadcast %r: %r\n" % (fun, exc))
            self.pause()

        return _insured(instance.broker.pool, fun, args, kwargs,
                        on_revive=state.on_broker_revive,
                        errback=errback)

    def ib(self, fun, *args, **kwargs):
        """Shortcut to ``self.insured(
            fun.im_self, fun(*args, **kwargs))``"""
        return self.insured(fun.im_self, fun, *args, **kwargs)

    def pause(self):
        self.paused = True  # only used in supervisor.

    def resume(self):
        self.paused = False

    def respond_to_ping(self):
        pass

    def _verify_restart_instance(self, instance):
        """Restarts the instance, and verifies that the instance is
        actually able to start."""
        self.info("%s instance.restart" % (instance, ))
        instance.restart()
        is_alive = False
        for i in fxrangemax(0.1, 1, 0.4, 30):
            self.info("%s pingWithTimeout: %s", instance, i)
            self.respond_to_ping()
            if self.insured(instance, instance.responds_to_ping, timeout=i):
                is_alive = True
                break
        if is_alive:
            self.info("%s successfully restarted" % (instance, ))
        else:
            self.info("%s instance doesn't respond after restart" % (
                    instance, ))

    def _can_restart(self):
        """Returns true if the supervisor is allowed to restart
        an instance at this point."""
        return True

    def _do_restart_instance(self, instance, ratelimit=False):
        bucket = self._buckets[instance.restart]
        if ratelimit:
            if self._can_restart():
                if bucket.can_consume(1):
                    self._verify_restart_instance(instance)
                else:
                    self.error(
                        "%s instance.disabled: Restarted too often", instance)
                    instance.disable()
                    self._buckets.pop(instance.restart)
        else:
            self._buckets.pop(instance.restart, None)
            self._verify_restart_instance(instance)

    def _do_stop_instance(self, instance):
        self.info("%s instance.shutdown" % (instance, ))
        instance.stop()

    def _do_stop_verify_instance(self, instance):
        self.info("%s instance.shutdown" % (instance, ))
        instance.stop_verify()

    def _do_verify_instance(self, instance, ratelimit=False):
        if not self.paused:
            if instance.is_enabled and instance.pk:
                if not self.ib(instance.alive):
                    self._do_restart_instance(instance, ratelimit=ratelimit)
                self._verify_instance_processes(instance)
                self._verify_instance_queues(instance)
            else:
                if self.ib(instance.alive):
                    self._do_stop_instance(instance)

    def _verify_instance_queues(self, instance):
        """Verify that the queues the instance is consuming from matches
        the queues listed in the model."""
        queues = set(instance.queues)
        reply = self.ib(instance.consuming_from)
        if reply is None:
            return
        consuming_from = set(reply.keys())

        for queue in consuming_from ^ queues:
            if queue in queues:
                self.info("%s: instance.consume_from: %s" % (instance, queue))
                self.ib(instance.add_queue, queue)
            elif queue == instance.direct_queue:
                pass
            else:
                self.info(
                    "%s: instance.cancel_consume: %s" % (instance, queue))
                self.ib(instance.cancel_queue, queue)

    def _verify_instance_processes(self, instance):
        """Verify that the max/min concurrency settings of the
        instance matches that which is specified in the model."""
        max, min = instance.max_concurrency, instance.min_concurrency
        try:
            current = self.insured(instance, instance.stats)["autoscaler"]
        except (TypeError, KeyError):
            return
        if max != current["max"] or min != current["min"]:
            self.info("%s: instance.set_autoscale max=%r min=%r" % (
                instance, max, min))
            self.ib(instance.autoscale, max, min)



"""cyme.api.views"""

from __future__ import absolute_import
from __future__ import with_statement

import re

from celery import current_app as celery
from celery.result import AsyncResult

from . import web
from ..branch.controller import apps, branches, instances, queues
from ..tasks import webhook
from ..utils import uuid


class Branch(web.ApiView):

    def get(self, request, branch=None):
        return branches.get(branch) if branch else branches.all()


class App(web.ApiView):

    def get(self, request, app=None):
        return apps.get(app).as_dict() if app else apps.all()

    def put(self, request, app=None):
        return self.Created(apps.add(app or uuid(),
                            **self.params("broker", "arguments",
                                          "extra_config")))
    post = put

    def delete(self, request, app):
        return apps.delete(app)


class Instance(web.ApiView):

    def get(self, request, app, name=None, nowait=False):
        return instances.get(name) if name else instances.all(app=app)

    def delete(self, request, app, name, nowait=False):
        return self.Ok(instances.remove(name, nowait=nowait))

    def post(self, request, app, name=None, nowait=False):
        return self.Created(instances.add(name=name, app=app,
                                      nowait=nowait,
                                      **self.params("broker", "pool",
                                                    "arguments",
                                                    "extra_config")))

    def put(self, *args, **kwargs):
        return self.NotImplemented("Operation is not idempotent: use POST")


class Consumer(web.ApiView):

    def get(self, request, app, name, queue=None, nowait=False):
        return instances.consuming_from(name)

    def put(self, request, app, name, queue, nowait=False):
        return self.Created(instances.add_consumer(name, queue, nowait=nowait))
    post = put

    def delete(self, request, app, name, queue, nowait=False):
        return self.Ok(instances.cancel_consumer(name, queue, nowait=nowait))


class Queue(web.ApiView):

    def get(self, request, app, name=None):
        return queues.get(name) if name else queues.all()

    def delete(self, request, app, name, nowait=False):
        return self.Ok(queues.delete(name))

    def put(self, request, app, name, nowait=False):
        return self.Created(queues.add(name, nowait=nowait,
                      **self.params("exchange", "exchange_type",
                                    "routing_key", "options")))
    post = put


class apply(web.ApiView):
    get_methods = frozenset(["GET", "HEAD"])
    re_find_queue = re.compile(r'/?(.+?)/?$')
    re_url_in_path = re.compile(r'(.+?/)(\w+://)(.+)')

    def prepare_path(self, rest):
        path, url = self._parse_path_containing_url(rest)
        if path:
            m = self.re_find_queue.match(path)
            if m:
                return m.groups()[0], url
        return None, url

    def dispatch(self, request, app, rest):
        gd = lambda m: getattr(request, m)
        queue, url = self.prepare_path(rest)
        app = apps.get(app)
        broker = app.get_broker()
        method = request.method.upper()
        pargs = {}
        if queue:
            queue = queues.get(queue)
            pargs.update(exchange=queue["exchange"],
                         exchange_type=queue["exchange_type"],
                         routing_key=queue["routing_key"])
        params = gd(method) if method in self.get_methods else gd("GET")
        data = gd(method) if method not in self.get_methods else None

        with broker.producers.acquire(block=True) as producer:
            publisher = celery.amqp.TaskPublisher(
                            connection=producer.connection,
                            channel=producer.channel)
            result = webhook.apply_async((url, method, params, data),
                                         publisher=publisher, retry=True,
                                         **pargs)
            return self.Accepted({"uuid": result.task_id, "url": url,
                                  "queue": queue, "method": method,
                                  "params": params, "data": data,
                                  "broker": producer.connection.as_uri()})

        def _parse_path_containing_url(self, rest):
            m = self.re_url_in_path.match(rest)
            if m:
                first, scheme, last = m.groups()
                if scheme:
                    return first, scheme + last
                return first, None
            return rest, None


class autoscale(web.ApiView):

    def get(self, request, app, name):
        instance = instances.get(name)
        return {"max": instance["max_concurrency"],
                "min": instance["min_concurrency"]}

    def post(self, request, app, name, nowait=False):
        return self.Ok(instances.autoscale(name, nowait=nowait,
                        **self.params(("max", int), ("min", int))))


@web.simple_get
def instance_stats(self, request, app, name):
    return instances.stats(name)


@web.simple_get
def task_state(self, request, app, uuid):
    return {"state": AsyncResult(uuid).state}


@web.simple_get
def task_result(self, request, app, uuid):
    return {"result": AsyncResult(uuid).result}


@web.simple_get
def task_wait(self, request, app, uuid):
    return {"result": AsyncResult(uuid).get()}


@web.simple_get
def ping(self, request):
    return {"ok": "pong"}



"""Since cyme works as a contained Django APP, this is the default settings
file used when cyme is used outside of a Django project context."""
from __future__ import absolute_import


import os
import djcelery
djcelery.setup_loader()

DEBUG = True

# Broker settings.
BROKER_HOST = "amqp://127.0.0.1:5672//"
BROKER_POOL_LIMIT = 100


CELERYD_LOG_FORMAT = """\
[%(asctime)s: %(levelname)s] %(message)s\
""".strip()

DB_NAME = os.environ.get("CYME_DB_NAME") or "branch.db"

# Databases
DATABASES = {'default': {
                'ENGINE': 'django.db.backends.sqlite3',
                'NAME': DB_NAME,
            }}

# URL and file paths.
SITE_ID = 1
STATIC_URL = "/static"
ADMIN_MEDIA_PREFIX = "/adminstatic/"
TEMPLATE_LOADERS = (
    ('django.template.loaders.cached.Loader', (
        'django.template.loaders.filesystem.Loader',
        'django.template.loaders.app_directories.Loader',
    )),
)
ROOT_URLCONF = "cyme.api.urls"

# Time and localization.
TIME_ZONE = "UTC"
LANGUAGE_CODE = "en-us"
USE_I18N = USE_L10N = True

# Apps and middleware.
INSTALLED_APPS = ("django.contrib.auth",
                  "django.contrib.contenttypes",
                  "django.contrib.sessions",
                  "cyme",  # cyme must be before admin.
                  "cyme.api",
                  "django.contrib.admin",
                  "django.contrib.admindocs")

MIDDLEWARE_CLASSES = (
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
)

# Make this unique, and don't share it with anybody.
SECRET_KEY = '9a3!m32h23psjjkkjl#()hs+-sv@$3*mgq!m3s!encow2&*738'



"""cyme.managers

- These are the managers for our models in :mod:`cyme.models`.

- They are not to be used directly, but accessed through
  the ``objects`` attribute of a Model.


"""

from __future__ import absolute_import

from anyjson import serialize
from celery import current_app as celery
from djcelery.managers import ExtendedManager

from ..utils import cached_property, uuid


class BrokerManager(ExtendedManager):

    def get_default(self):
        return self.get_or_create(url=self.default_url)[0]

    @property
    def default_url(self):
        return celery.broker_connection().as_uri()


class AppManager(ExtendedManager):

    def from_json(self, name=None, broker=None):
        return {"name": name, "broker": self.get_broker(broker)}

    def recreate(self, name=None, broker=None, arguments=None,
            extra_config=None):
        d = self.from_json(name, broker)
        return self.get_or_create(name=d["name"],
                                  defaults={"broker": d["broker"],
                                            "arguments": arguments,
                                            "extra_config": extra_config})[0]

    def instance(self, name=None, broker=None):
        return self.model(**self.from_json(name, broker))

    def get_broker(self, url):
        return self.Brokers.get_or_create(url=url)[0]

    def add(self, name=None, broker=None, arguments=None, extra_config=None):
        broker = self.get_broker(broker) if broker else None
        return self.get_or_create(name=name, defaults={
                "broker": broker,
                "arguments": arguments,
                "extra_config": extra_config})[0]

    def get_default(self):
        return self.get_or_create(name="cyme")[0]

    @cached_property
    def Brokers(self):
        return self.model.Broker._default_manager


class InstanceManager(ExtendedManager):

    def enabled(self):
        return self.filter(is_enabled=True)

    def _maybe_queues(self, queues):
        if isinstance(queues, basestring):
            queues = queues.split(",")
        return [(queue.name if isinstance(queue, self.model.Queue) else queue)
                    for queue in queues]

    def add(self, name=None, queues=None, max_concurrency=1,
            min_concurrency=1, broker=None, pool=None, app=None,
            arguments=None, extra_config=None):
        instance = self.create(name=name or uuid(),
                               max_concurrency=max_concurrency,
                               min_concurrency=min_concurrency,
                               pool=pool,
                               app=app,
                               arguments=arguments,
                               extra_config=extra_config)
        needs_save = False
        if queues:
            instance.queues = self._maybe_queues(queues)
            needs_save = True
        if broker:
            instance._broker = broker
            needs_save = True
        if needs_save:
            instance.save()
        return instance

    def _action(self, name, action, *args, **kwargs):
        instance = self.get(name=name)
        getattr(instance, action)(*args, **kwargs)
        return instance

    def remove(self, name):
        return self._action(name, "delete")

    def enable(self, name):
        return self._action(name, "enable")

    def disable(self, name):
        return self._action(name, "disable")

    def remove_queue_from_instances(self, queue, **query):
        instances = []
        for instance in self.filter(**query).iterator():
            if queue in instance.queues:
                instance.queues.remove(queue)
                instance.save()
                instances.append(instance)
        return instances

    def add_queue_to_instances(self, queue, **query):
        instances = []
        for instance in self.filter(**query).iterator():
            instance.queues.add(queue)
            instance.save()
            instances.append(instance)
        return instances


class QueueManager(ExtendedManager):

    def enabled(self):
        return self.filter(is_enabled=True)

    def _add(self, name, **declaration):
        return self.get_or_create(name=name, defaults=declaration)[0]

    def add(self, name, exchange=None, exchange_type=None,
            routing_key=None, **options):
        options = serialize(options) if options else None
        return self._add(name, exchange=exchange, exchange_type=exchange_type,
                               routing_key=routing_key, options=options)



def setup(app):
    app.add_crossref_type(
        directivename="setting",
        rolename="setting",
        indextemplate="pair: %s; setting",
    )
    app.add_crossref_type(
        directivename="sig",
        rolename="sig",
        indextemplate="pair: %s; sig",
    )
    app.add_crossref_type(
        directivename="state",
        rolename="state",
        indextemplate="pair: %s; state",
    )
    app.add_crossref_type(
        directivename="control",
        rolename="control",
        indextemplate="pair: %s; control",
    )
    app.add_crossref_type(
        directivename="signal",
        rolename="signal",
        indextemplate="pair: %s; signal",
    )



"""cyme.branch.signals"""

from __future__ import absolute_import

from celery.utils.dispatch import Signal

#: Sent when the http server is ready to accept requests.
#: Arguments:
#:
#:     :sender: the :class:`~cyme.httpd.HttpServer` instance.
#:     :addrport: the ``(hostname, port)`` tuple.
#:     :handler: the WSGI handler used.
#:     :sock: the socket used.
httpd_ready = Signal(providing_args=["addrport", "handler", "sock"])

#: Sent when the supervisor is ready.
#: Arguments:
#:
#:     :sender: is the :class:`~cyme.supervisor.Supervisor` instance.
supervisor_ready = Signal()

#: Sent when a controller is ready.
#:
#: Arguments:
#:     :sender: is the :class:`~cyme.controller.Controller` instance.
controller_ready = Signal()

#: Sent when the branch and all its components are ready to serve.
#:
#: Arguments:
#:     :sender: is the :class:`~cyme.branch.Branch` instance.
branch_ready = Signal()

branch_shutdown_complete = Signal()


branch_startup_request = Signal()
branch_shutdown_request = Signal()


thread_pre_shutdown = Signal()
thread_pre_join = Signal(providing_args=["timeout"])
thread_exit = Signal()
thread_post_join = Signal()
thread_post_shutdown = Signal()
thread_shutdown_step = Signal()

thread_pre_start = Signal()
thread_post_start = Signal()
thread_startup_step = Signal()
presence_ready = Signal()



"""cyme.branch.supervisor"""

from __future__ import absolute_import
from __future__ import with_statement

from threading import Lock
from Queue import Empty

from celery.local import Proxy
from eventlet.queue import LightQueue
from eventlet.event import Event

from .signals import supervisor_ready
from .thread import gThread

from ..status import Status

__current = None


class Supervisor(gThread, Status):
    """The supervisor wakes up at intervals to monitor changes in the model.
    It can also be requested to perform specific operations, and these
    operations can be either async or sync.

    :keyword interval:  This is the interval (in seconds as an int/float),
       between verifying all the registered instances.
    :keyword queue: Custom :class:`~Queue.Queue` instance used to send
        and receive commands.

    It is responsible for:

        * Stopping removed instances.
        * Starting new instances.
        * Restarting unresponsive/killed instances.
        * Making sure the instances consumes from the queues
          specified in the model, sending ``add_consumer``/-
          ``cancel_consumer`` broadcast commands to the instances as it
          finds inconsistencies.
        * Making sure the max/min concurrency setting is as specified in the
          model,  sending ``autoscale`` broadcast commands to the noes
          as it finds inconsistencies.

    The supervisor is resilient to intermittent connection failures,
    and will auto-retry any operation that is dependent on a broker.

    Since workers cannot respond to broadcast commands while the
    broker is off-line, the supervisor will not restart affected
    instances until the instance has had a chance to reconnect (decided
    by the :attr:`wait_after_broker_revived` attribute).

    """
    #: Limit instance restarts to 1/m, so out of control
    #: instances will be disabled
    restart_max_rate = "1/m"

    #: Default interval_max for ensure_connection is 30 secs.
    wait_after_broker_revived = 35.0

    #: Connection errors pauses the supervisor, so events does not accumulate.
    paused = False

    #: Default interval (time in seconds as a float to reschedule).
    interval = 60.0

    def __init__(self, interval=None, queue=None, set_as_current=True):
        self.set_as_current = set_as_current
        if self.set_as_current:
            set_current(self)
        self._orig_queue_arg = queue
        self.interval = interval or self.interval
        self.queue = LightQueue() if queue is None else queue
        self._pause_mutex = Lock()
        self._last_update = None
        gThread.__init__(self)
        Status.__init__(self)

    def __copy__(self):
        return self.__class__(self.interval, self._orig_queue_arg)

    def pause(self):
        """Pause all timers."""
        self.respond_to_ping()
        with self._pause_mutex:
            if not self.paused:
                self.debug("pausing")
                self.paused = True

    def resume(self):
        """Resume all timers."""
        with self._pause_mutex:
            if self.paused:
                self.debug("resuming")
                self.paused = False

    def verify(self, instances, ratelimit=False):
        """Verify the consistency of one or more instances.

        :param instances: List of instances to verify.

        This operation is asynchronous, and returns a :class:`Greenlet`
        instance that can be used to wait for the operation to complete.

        """
        return self._request(instances, self._do_verify_instance,
                            {"ratelimit": ratelimit})

    def restart(self, instances):
        """Restart one or more instances.

        :param instances: List of instances to restart.

        This operation is asynchronous, and returns a :class:`Greenlet`
        instance that can be used to wait for the operation to complete.

        """
        return self._request(instances, self._do_restart_instance)

    def shutdown(self, instances):
        """Shutdown one or more instances.

        :param instances: List of instances to stop.

        This operation is asynchronous, and returns a :class:`Greenlet`
        instance that can be used to wait for the operation to complete.

        .. warning::

            Note that the supervisor will automatically restart
            any stopped instances unless the corresponding :class:`Instance`
            model has been marked as disabled.

        """
        return self._request(instances, self._do_stop_instance)

    def _request(self, instances, action, kwargs={}):
        event = Event()
        self.queue.put_nowait((instances, event, action, kwargs))
        return event

    def before(self):
        self.start_periodic_timer(self.interval, self._verify_all)

    def run(self):
        queue = self.queue
        self.info("started")
        supervisor_ready.send(sender=self)
        while not self.should_stop:
            try:
                instances, event, action, kwargs = queue.get(timeout=1)
            except Empty:
                self.respond_to_ping()
                continue
            self.respond_to_ping()
            self.debug("wake-up")
            try:
                for instance in instances:
                    try:
                        action(instance, **kwargs)
                    except Exception, exc:
                        self.error("Event caused exception: %r", exc)
            finally:
                event.send(True)

    def _verify_all(self, force=False):
        if self._last_update and self._last_update.ready():
            try:
                self._last_update.wait()  # collect result
            except self.GreenletExit:
                pass
            force = True
        if not self._last_update or force:
            self._last_update = self.verify(self.all_instances(),
                                            ratelimit=True)


class _OfflineSupervisor(object):

    def wait(self):
        pass

    def _noop(self, *args, **kwargs):
        return self
    pause = resume = verify = shutdown = restart = _noop


def set_current(sup):
    global __current
    __current = sup
    return __current


def get_current():
    if __current is None:
        return _OfflineSupervisor()
    return __current

supervisor = Proxy(get_current)



"""cyme.branch.managers

- Contains the :class:`LocalInstanceManager` instance,
  which is the preferred API used to control and manage worker
  instances handled by this branch.  I.e. it can be used to do synchronous
  actions that don't return until the supervisor has performed them.

"""

from __future__ import absolute_import

from .supervisor import supervisor as sup

from ..models import Broker, Instance
from ..utils import force_list


class LocalInstanceManager(object):
    Instances = Instance._default_manager
    Brokers = Broker._default_manager

    def get(self, name):
        return self.Instances.get(name=name)

    def add(self, name=None, queues=None,
            max_concurrency=1, min_concurrency=1, broker=None,
            pool=None, app=None, arguments=None, extra_config=None,
            nowait=False, **kwargs):
        broker = self.Brokers.get_or_create(url=broker)[0] if broker else None
        instance = self.Instances.add(name, queues, max_concurrency,
                                  min_concurrency, broker, pool, app,
                                  arguments, extra_config)
        return self.maybe_wait(sup.verify, instance, nowait)

    def remove(self, name, nowait=False):
        return self.maybe_wait(sup.shutdown,
                               self.Instances.remove(name), nowait)

    def restart(self, name, nowait=False):
        return self.maybe_wait(sup.restart, self.get(name), nowait)

    def enable(self, name, nowait=False):
        return self.maybe_wait(sup.verify,
                               self.Instances.enable(name), nowait)

    def disable(self, name, nowait=False):
        return self.maybe_wait(sup.verify,
                               self.Instances.disable(name), nowait)

    def add_consumer(self, name, queue, nowait=False):
        return self.maybe_wait(sup.verify,
                self.get(name).add_queue_eventually(queue), nowait)

    def cancel_consumer(self, name, queue, nowait=False):
        return self.maybe_wait(sup.verify,
                self.Instances.remove_queue_from_instances(queue, name=name),
                nowait)

    def remove_queue(self, queue, nowait=False):
        return self.maybe_wait(sup.verify,
                self.Instances.remove_queue_from_instances(queue), nowait)

    def maybe_wait(self, fun, instances, nowait):
        if instances:
            g = fun(force_list(instances))
            nowait and g.wait()
        return instances
local_instances = LocalInstanceManager()



"""cyme.branch

- This is the Branch thread started by the :program:`cyme-branch` program.

  It starts the HTTP server, the Supervisor, and one or more controllers.

"""

from __future__ import absolute_import

import logging

from celery import current_app as celery
from celery.utils import LOG_LEVELS, term
from cl.g import Event
from kombu.log import LogMixin
from kombu.utils import gen_unique_id

from . import signals
from .state import state
from .thread import gThread

from ..utils import find_symbol, instantiate


class MockSup(LogMixin):

    def __init__(self, thread, *args):
        self.thread = thread

    def start(self):
        self.thread.start()

    def stop(self):
        return self.thread.stop()


class Branch(gThread):
    controller_cls = ".controller.Controller"
    httpd_cls = ".httpd.HttpServer"
    supervisor_cls = ".supervisor.Supervisor"
    intsup_cls = ".intsup.gSup"

    _components_ready = {}
    _components_shutdown = {}
    _presence_ready = 0
    _ready = False

    def __init__(self, addrport="", id=None, loglevel=logging.INFO,
            logfile=None, without_httpd=False, numc=2, sup_interval=None,
            ready_event=None, colored=None, **kwargs):
        self.id = id or gen_unique_id()
        if isinstance(addrport, basestring):
            addr, _, port = addrport.partition(":")
            addrport = (addr, int(port) if port else 8000)
        self.addrport = addrport
        self.connection = celery.broker_connection()
        self.without_httpd = without_httpd
        self.logfile = logfile
        self.loglevel = loglevel
        self.numc = numc
        self.ready_event = ready_event
        self.exit_request = Event()
        self.colored = colored or term.colored(enabled=False)
        self.httpd = None
        gSup = find_symbol(self, self.intsup_cls)
        if not self.without_httpd:
            self.httpd = MockSup(instantiate(self, self.httpd_cls, addrport),
                              signals.httpd_ready)
        self.supervisor = gSup(instantiate(self, self.supervisor_cls,
                                sup_interval), signals.supervisor_ready)
        self.controllers = [gSup(instantiate(self, self.controller_cls,
                                   id="%s.%s" % (self.id, i),
                                   connection=self.connection,
                                   branch=self),
                                 signals.controller_ready)
                                for i in xrange(1, numc + 1)]
        c = [self.supervisor] + self.controllers + [self.httpd]
        c = self.components = list(filter(None, c))
        self._components_ready = dict(zip([z.thread for z in c],
                                          [False] * len(c)))
        for controller in self.controllers:
            if hasattr(controller.thread, "presence"):
                self._components_ready[controller.thread.presence] = False
        self._components_shutdown = dict(self._components_ready)
        super(Branch, self).__init__()

    def _component_ready(self, sender=None, **kwargs):
        if not self._ready:
            self._components_ready[sender] = True
            if all(self._components_ready.values()):
                signals.branch_ready.send(sender=self)
                if self.ready_event:
                    self.ready_event.send()
                    self.ready_event = None
                self._ready = True

    def on_ready(self, **kwargs):
        pass

    def prepare_signals(self):
        signals.controller_ready.connect(self._component_ready)
        signals.httpd_ready.connect(self._component_ready)
        signals.supervisor_ready.connect(self._component_ready)
        signals.presence_ready.connect(self._component_ready)
        signals.branch_ready.connect(self.on_ready)
        signals.thread_post_shutdown.connect(self._component_shutdown)

    def run(self):
        state.is_branch = True
        signals.branch_startup_request.send(sender=self)
        self.prepare_signals()
        self.info("Starting with id %r", self.id)
        [g.start() for g in self.components]
        self.exit_request.wait()

    def stop(self):
        self.exit_request.send(1)
        super(Branch, self).stop()

    def after(self):
        for component in reversed(self.components):
            if self._components_ready[component.thread]:
                try:
                    component.stop()
                except KeyboardInterrupt:
                    pass
                except BaseException, exc:
                    component.error("Error in shutdown: %r", exc)

    def _component_shutdown(self, sender, **kwargs):
        self._components_shutdown[sender] = True
        if all(self._components_shutdown.values()):
            signals.branch_shutdown_complete.send(sender=self)

    def about(self):
        url = port = None
        if self.httpd:
            url, port = self.http.thread.url, self.httpd.thread.port
        port = self.httpd.thread.port if self.httpd else None
        return {"id": self.id,
                "loglevel": LOG_LEVELS[self.loglevel],
                "numc": self.numc,
                "sup_interval": self.supervisor.interval,
                "logfile": self.logfile,
                "port": port,
                "url": url}



"""cyme.api.urls"""

from __future__ import absolute_import

from django.contrib import admin
from django.conf.urls.defaults import (patterns, include, url,  # noqa
                                       handler500, handler404)

from . import views

uApp = r'(?P<app>[^/]+)'
uNowait = r'(?P<nowait>!/)?'

admin.autodiscover()


def _o_(u):
    return (u.replace("APP", uApp)
             .replace("!", uNowait))

urlpatterns = patterns('',
    (r'^ping/$', views.ping.as_view()),
    (r'^admin/doc/', include('django.contrib.admindocs.urls')),

    (r'^admin/', include(admin.site.urls)),
    (r'^branches/(?P<branch>.+?)?/?$', views.Branch.as_view()),
    (_o_(r'^APP/queue/!(?P<rest>.+)'), views.apply.as_view()),
    (_o_(r'^APP/queues/!/?$'), views.Queue.as_view()),
    (_o_(r'^APP/queues/!(?P<name>.+?)/?$'), views.Queue.as_view()),
    (_o_(r'^APP/instances/!(?P<name>.+?)/queues/(?P<queue>.+?)?/?$'),
        views.Consumer.as_view()),
    (_o_(r'^APP/instances/!?(?P<name>.+)?/autoscale/?'),
        views.autoscale.as_view()),
    (_o_(r'^APP/instances/!(?P<name>.+)?/stats/?'),
        views.instance_stats.as_view()),
    (_o_(r'^APP/instances/!(?P<name>.+?)?/?$'), views.Instance.as_view()),
    (_o_(r'^APP/query/(?P<uuid>.+?)/state/?'), views.task_state.as_view()),
    (_o_(r'^APP/query/(?P<uuid>.+?)/result/?'), views.task_result.as_view()),
    (_o_(r'^APP/query/(?P<uuid>.+?)/wait/?'), views.task_wait.as_view()),
    (_o_(r'^APP?/?$'), views.App.as_view()),
)



"""Cyme - Celery instance manager"""
from __future__ import absolute_import

import os

VERSION = (0, 0, 3)

__version__ = ".".join(map(str, VERSION[0:3])) + "".join(VERSION[3:])
__author__ = "Ask Solem"
__contact__ = "ask@celeryproject.org"
__homepage__ = "http://celeryproject.org"
__docformat__ = "restructuredtext"
__license__ = "BSD (3 clause)"

DEBUG = os.environ.get("CYME_DEBUG")
DEBUG_BLOCK = os.environ.get("CYME_DEBUG_BLOCK")
DEBUG_READERS = os.environ.get("CYME_DEBUG_READERS")
if not os.environ.get("CYME_NO_EVAL"):
    from .client import Client  # noqa



"""cyme.tasks

- We have our own version of the webhook task.

- It simply forwards the original request, not depending on any semantics
  present in the query string of the request, nor in the data returned
  in the response.

"""
from __future__ import absolute_import

from celery.task import task
from requests import request

from . import __version__

#: Cyme User Agent string.
UA = "Celery/cyme v%s" % (__version__, )

#: Default HTTP headers to pass to the dispatched request.
DEFAULT_HEADERS = {"User-Agent": UA}


def response_to_dict(r):
    return dict(status_code=r.status_code, url=r.url,
                headers=r.headers, content=r.read())


@task(timeout=60)
def webhook(url, method="GET", params={}, data={}, headers=None, **kwargs):
    kwargs["timeout"] = kwargs.get("timeout", webhook.timeout)
    headers = {} if headers is None else headers
    return response_to_dict(request(method, url, params=params, data=data,
                                    headers=dict(headers, **DEFAULT_HEADERS)))



"""cyme.branch.controller

- Actors used to manage entities across all branches.

"""

from __future__ import absolute_import

from functools import partial

from celery.actors import Actor, AwareAgent
from cl.presence import AwareActorMixin, announce_after
from cl.utils import flatten, first_or_raise, shortuuid
from celery import current_app as celery
from kombu import Exchange
from kombu.common import uuid

from . import metrics
from . import signals
from .state import state
from .thread import gThread

from .. import conf
from .. import models
from ..utils import cached_property, find_symbol, promise


class CymeActor(Actor, AwareActorMixin):
    _announced = set()  # note: global

    def setup(self):
        # retry publishing messages by default if running as cyme-branch.
        self.retry = state.is_branch
        self.default_fields = {"actor_id": self.id}


class ModelActor(CymeActor):
    model = None

    def on_agent_ready(self):
        if self.name not in self._announced:
            self.log.info("%s: %s", self.name_plural,
                    promise(lambda: ", ".join(self.state.all())))
            self._announced.add(self.name)

    def contribute_to_state(self, state):
        state.model = self.model
        state.objects = self.model._default_manager
        return Actor.contribute_to_state(self, state)

    @cached_property
    def name(self):
        return unicode(self.model._meta.verbose_name.capitalize())

    @cached_property
    def name_plural(self):
        return unicode(self.model._meta.verbose_name_plural).capitalize()


class Branch(CymeActor):
    exchange = Exchange("cyme.Branch")
    default_timeout = 60
    types = ("direct", "scatter", "round-robin")
    meta_lookup_section = "this"

    def on_agent_ready(self):
        if self.name not in self._announced:
            self.log.info("Actor ready", self.name)
        self._announced.add(self.name)

    class state:

        def id(self):
            return self.agent.branch.id

        def url(self):
            return self.agent.branch.httpd.thread.url

        def about(self):
            return self.agent.branch.about()

        def shutdown(self, id):
            if id in [self.id(), "*"]:
                assert state.is_branch
                self.log.warn("Shutdown requested from remote.")
                raise SystemExit()
            raise self.Next()

    def all(self, **kw):
        return flatten(self.scatter("id", **kw))

    def get(self, id, **kw):
        return self.send_to_able("about", to=id, **kw)

    def url(self, id=None, **kw):
        if id:
            return self.send_to_able("url", to=id, **kw)
        return flatten(self.scatter("url", **kw))

    def shutdown(self, id):
        return self.send_to_able("shutdown", {"id": id}, to=id, nowait=True)

    def shutdown_all(self):
        return self.scatter("shutdown", {"id": "*"}, nowait=True)

    @property
    def meta(self):
        return {"this": [self.state.id()]}
branches = Branch()


class App(ModelActor):
    """Actor for managing the app model."""
    model = models.App
    types = ("scatter", )
    exchange = Exchange("cyme.App")
    _cache = {}

    class state:

        def all(self):
            return [app.name for app in self.objects.all()]

        def add(self, name, broker=None, arguments=None, extra_config=None):
            return self.objects.add(name, broker=broker,
                                          arguments=arguments,
                                          extra_config=extra_config).as_dict()

        def delete(self, name):
            return self.objects.filter(name=name).delete() and "ok"

        def get(self, name):
            try:
                return self.objects.get(name=name).as_dict()
            except self.model.DoesNotExist:
                raise self.Next()

        def metrics(self):
            instance_dir = str(conf.CYME_INSTANCE_DIR)
            return {"load_average": metrics.load_average(),
                    "disk_use": metrics.df(instance_dir).capacity}

    def all(self):
        return flatten(self.scatter("all"))

    def add(self, name, **broker):
        self.scatter("add", dict({"name": name}, **broker), nowait=True)
        return self.state.add(name, **broker)

    def delete(self, name, **kw):
        self._cache.pop(name, None)
        return list(self.scatter("delete", dict({"name": name}, **kw)))

    def metrics(self, name=None):
        return list(self.scatter("metrics"))

    def get(self, name=None):
        objects = self.state.objects
        if not name:
            return objects.get_default()
        if name not in self._cache:
            app = self._get(name)
            if not app:
                raise KeyError(name)
            self._cache[name] = objects.recreate(**app)

        return self._cache[name]

    def _get(self, name):
        try:
            return self.state.get(name)
        except self.Next:
            return first_or_raise(self.scatter("get", {"name": name},
                                               propagate=False),
                                  self.NoRouteError(name))
apps = App()


class Instance(ModelActor):
    """Actor for managing the Instance model."""
    model = models.Instance
    exchange = Exchange("cyme.Instance")
    default_timeout = 60
    types = ("direct", "scatter", "round-robin")
    meta_lookup_section = "instances"

    class state:

        def all(self, app=None):
            fun = self.objects.all
            if app:
                fun = partial(self.objects.filter, app=apps.get(app))
            return [instance.name for instance in fun()]

        def get(self, name, app=None):
            try:
                x = self.objects.get(name=name)
            except self.model.DoesNotExist:
                raise self.Next()
            return x.as_dict()

        @announce_after
        def add(self, name=None, app=None, **kwargs):
            return self.local.add(name, app=apps.get(app), **kwargs).as_dict()

        @announce_after
        def remove(self, name, app=None):
            return self.local.remove(name) and "ok"

        def restart(self, name, app=None):
            return self.local.restart(name) and "ok"

        def enable(self, name, app=None):
            return self.local.enable(name) and "ok"

        def disable(self, name, app=None):
            return self.local.disable(name) and "ok"

        def add_consumer(self, name, queue):
            return self.local.add_consumer(name, queue) and "ok"

        def cancel_consumer(self, name, queue):
            return self.local.cancel_consumer(name, queue) and "ok"

        def remove_queue_from_all(self, queue):
            return [instance.name for instance in
                        self.objects.remove_queue_from_instances(queue)]

        def autoscale(self, name, max=None, min=None):
            instance = self.local.get(name)
            instance.autoscale(max=max, min=min)
            return {"max": instance.max_concurrency,
                    "min": instance.min_concurrency}

        def consuming_from(self, name):
            return self.local.get(name).consuming_from()

        def stats(self, name):
            return self.local.get(name).stats()

        @cached_property
        def local(self):
            return find_symbol(self, ".managers.local_instances")

    def get(self, name, app=None, **kw):
        return self.send_to_able("get",
                                 {"name": name, "app": app}, to=name, **kw)

    def all(self, app=None):
        return flatten(self.scatter("all", {"app": app}))

    def add(self, name=None, app=None, nowait=False, **kwargs):
        if nowait:
            name = name if name else uuid()
        ret = self.throw("add", dict({"name": name, "app": app}, **kwargs),
                         nowait=nowait)
        if nowait:
            return {"name": name}
        return ret

    def remove(self, name, **kw):
        return self.send_to_able("remove", {"name": name}, to=name, **kw)

    def restart(self, name, **kw):
        return self.send_to_able("restart", {"name": name}, to=name, **kw)

    def enable(self, name, **kw):
        return self.send_to_able("enable", args={"name": name}, to=name, **kw)

    def disable(self, name, **kw):
        return self.send_to_able("disable", args={"name": name}, to=name, **kw)

    def add_consumer(self, name, queue, **kw):
        return self.send_to_able("add_consumer",
                                 {"name": name, "queue": queue}, to=name, **kw)

    def cancel_consumer(self, name, queue, **kw):
        return self.send_to_able("cancel_consumer",
                                 {"name": name, "queue": queue}, to=name, **kw)

    def remove_queue_from_all(self, queue, **kw):
        return flatten(self.scatter("remove_queue_from_all",
                                    {"queue": queue}, **kw) or [])

    def autoscale(self, name, max=None, min=None, **kw):
        return self.send_to_able("autoscale",
                         {"name": name, "min": min, "max": max}, to=name, **kw)

    def consuming_from(self, name, **kw):
        return self.send_to_able("consuming_from",
                                 {"name": name}, to=name, **kw)

    def stats(self, name, **kw):
        return self.send_to_able("stats", {"name": name}, to=name, **kw)

    @property
    def meta(self):
        return {"instances": self.state.all()}
instances = Instance()


class Queue(ModelActor):
    """Actor for managing the Queue model."""
    model = models.Queue
    exchange = Exchange("cyme.Queue")
    types = ("direct", "scatter", "round-robin")
    default_timeout = 2
    meta_lookup_section = "queues"

    class state:

        def all(self):
            return [q.name for q in self.objects.all()]

        def get(self, name):
            try:
                return self.objects.get(name=name).as_dict()
            except self.model.DoesNotExist:
                raise KeyError(name)

        @announce_after
        def add(self, name, **declaration):
            return self.objects._add(name, **declaration).as_dict()

        @announce_after
        def delete(self, name):
            self.objects.filter(name=name).delete()
            return "ok"

    def all(self):
        return flatten(self.scatter("all"))

    def get(self, name):
        try:
            # see if we have the queue locally.
            return self.state.get(name)
        except KeyError:
            # if not, ask the agents.
            return self.send_to_able("get", {"name": name}, to=name)

    def add(self, name, nowait=False, **decl):
        return self.throw("add", dict({"name": name}, **decl), nowait=nowait)

    def delete(self, name, **kw):
        instances.remove_queue_from_all(name, nowait=True)
        return self.send_to_able("delete", {"name": name}, to=name, **kw)

    @property
    def meta(self):
        return {"queues": self.state.all()}
queues = Queue()


class Controller(AwareAgent, gThread):
    actors = [Branch(), App(), Instance(), Queue()]
    connect_max_retries = celery.conf.BROKER_CONNECTION_MAX_RETRIES
    extra_shutdown_steps = 2
    _ready_sent = False
    _presence_ready_sent = False

    def __init__(self, *args, **kwargs):
        self.branch = kwargs.pop("branch", None)
        AwareAgent.__init__(self, *args, **kwargs)
        gThread.__init__(self)

    def on_awake(self):
        # bind global actors to this agent,
        # so presence can be used.
        for actor in (branches, apps, instances, queues):
            actor.agent = self

    def on_connection_revived(self):
        state.on_broker_revive()

    def on_consume_ready(self, *args, **kwargs):
        if not self._ready_sent:
            signals.controller_ready.send(sender=self)
            self._ready_sent = True
        super(Controller, self).on_consume_ready()

    def on_iteration(self):
        self.respond_to_ping()

    def on_connection_error(self, exc, interval):
        self.respond_to_ping()
        super(Controller, self).on_connection_error(exc, interval)

    def on_presence_ready(self):
        if not self._presence_ready_sent:
            signals.presence_ready.send(sender=self.presence)
            self._presence_ready_sent = True

    def stop(self):
        self.should_stop = True
        if hasattr(self, "presence") and self.presence.g:
            self.debug("waiting for presence to exit")
            signals.thread_shutdown_step.send(sender=self)
            self.presence.g.wait()
            signals.thread_shutdown_step.send(sender=self)
        super(Controller, self).stop()

    @property
    def logger_name(self):
        return '#'.join([self.__class__.__name__, self._shortid()])

    def _shortid(self):
        if '.' in self.id:
            return shortuuid(self.id) + ".." + self.id[-2:]
        return shortuuid(self.id)



"""cyme.bin.base

- Utilities used by command line applications,
  basically just to set up Django without having an actual project.

"""

from __future__ import absolute_import
from __future__ import with_statement

from .. import __version__, DEBUG, DEBUG_BLOCK, DEBUG_READERS

import getpass
import sys

from kombu.utils import cached_property

from ..utils import Path


class Env(object):

    def __init__(self, needs_eventlet=False, instance_dir=None):
        self.needs_eventlet = needs_eventlet
        self.instance_dir = instance_dir

    def __enter__(self):
        if self.needs_eventlet:
            self.setup_eventlet()
        self.configure()
        self.setup_pool_limit()
        return self

    def __exit__(self, *exc_info):
        pass

    def setup_eventlet(self):
        import eventlet
        import eventlet.debug
        eventlet.monkey_patch()
        if DEBUG_READERS:
            eventlet.debug.hub_prevent_multiple_readers(False)
            print("+++ MULTIPLE READERS ALLOWED +++")    # noqa+
        if DEBUG_BLOCK:
            eventlet.debug.hub_blocking_detection(True)
            print("+++ BLOCKING DETECTION ENABLED +++")  # noqa+

    def configure(self):
        from django.conf import settings
        from .. import settings as default_settings
        from ..utils import imerge_settings

        if not settings.configured:
            self.management.setup_environ(default_settings)
        else:
            imerge_settings(settings, default_settings)
        if self.instance_dir:
            settings.CYME_INSTANCE_DIR = self.instance_dir

    def setup_pool_limit(self, **kwargs):
        from kombu import pools
        from celery import current_app as celery
        limit = kwargs.get("limit", celery.conf.BROKER_POOL_LIMIT)
        pools.set_limit(limit if self.needs_eventlet else 1)
        celery._pool = pools.connections[celery.broker_connection()]

    def syncdb(self, interactive=True):
        from django.conf import settings
        from django.db.utils import DEFAULT_DB_ALIAS
        dbconf = settings.DATABASES[DEFAULT_DB_ALIAS]
        if dbconf["ENGINE"] == "django.db.backends.sqlite3":
            if Path(dbconf["NAME"]).absolute().exists():
                return
        gp, getpass.getpass = getpass.getpass, getpass.fallback_getpass
        try:
            self.management.call_command("syncdb", interactive=interactive)
        finally:
            getpass.getpass = gp

    @cached_property
    def management(self):
        from django.core import management
        return management


class BaseApp(object):
    env = None
    needs_eventlet = False
    instance_dir = None

    def get_version(self):
        return "cyme v%s" % (__version__, )

    def run_from_argv(self, argv=None):
        argv = sys.argv if argv is None else argv
        if '--version' in argv:
            print(self.get_version())
            sys.exit(0)
        try:
            with (self.env or
                    Env(self.needs_eventlet, self.instance_dir)) as env:
                return self.run(env, argv)
        except KeyboardInterrupt:
            if DEBUG:
                raise
            raise SystemExit()
    __call__ = run_from_argv


def app(**attrs):
    x = attrs

    def _app(fun):

        def run(self, *args, **kwargs):
            return fun(*args, **kwargs)

        attrs = dict({"run": run, "__module__": fun.__module__,
                      "__doc__": fun.__doc__}, **x)

        return type(fun.__name__, (BaseApp, ), attrs)()

    return _app



"""cyme.api.web

- Contains utilities for creating our HTTP API.

"""

from __future__ import absolute_import

import httplib as http
import sys

from functools import partial
from traceback import format_exception

from django.http import HttpResponse, HttpResponseNotFound
from django.views.generic.base import View

from anyjson import serialize
from cl.exceptions import NoReplyError, NoRouteError
from kombu.utils.encoding import safe_repr


class HttpResponseTimeout(HttpResponse):
    """The operation timed out."""
    status_code = http.REQUEST_TIMEOUT


class HttpResponseNotImplemented(HttpResponse):
    """The requested action is not implemented.
    Used for async requests when the operation is inherently sync."""
    status_code = http.NOT_IMPLEMENTED


def JsonResponse(data, status=http.OK, **kwargs):
    """Returns a JSON encoded response."""
    if isinstance(data, (basestring, int, float, bool)):
        data = {"ok": data}
    if data is None or not isinstance(data, (dict, list, tuple)):
        return data
    kwargs.setdefault("content_type", "application/json")
    response = HttpResponse(serialize(data),
                            status=status, **kwargs)
    response.csrf_exempt = True
    return response
Accepted = partial(JsonResponse, status=http.ACCEPTED)
Created = partial(JsonResponse, status=http.CREATED)
Error = partial(JsonResponse, status=http.INTERNAL_SERVER_ERROR)


class ApiView(View):
    nowait = False  # should the current operation be async?
    typemap = {int: lambda i: int(i) if i else None,
               float: lambda f: float(f) if f else None}
    _semipredicate = object()

    def dispatch(self, request, *args, **kwargs):
        self.nowait = kwargs.get("nowait", False)
        if request.method.lower() == "get":
            kwargs.pop("nowait", None)
            if self.nowait:
                return self.NotImplemented("Operation can't be async.")
        try:
            data = super(ApiView, self).dispatch(request, *args, **kwargs)
        except NoRouteError:
            return HttpResponseNotFound()
        except NoReplyError:
            return HttpResponseTimeout()
        except Exception, exc:
            return Error({"nok": [safe_repr(exc),
                                  "".join(format_exception(*sys.exc_info()))]})
        return self.Response(data)

    def Response(self, *args, **kwargs):
        return JsonResponse(*args, **kwargs)

    def Accepted(self, *args, **kwargs):
        return Accepted(*args, **kwargs)

    def Ok(self, data, *args, **kwargs):
        if self.nowait:
            data = data or {"ok": "operation scheduled"}
            return self.Accepted(data, **kwargs)
        return self.Response(data, *args, **kwargs)

    def Created(self, data, *args, **kwargs):
        if self.nowait:
            data = data or {"ok": "operation scheduled"}
            return self.Accepted(data, **kwargs)
        return Created(data, *args, **kwargs)

    def NotImplemented(self, *args, **kwargs):
        return HttpResponseNotImplemented(*args, **kwargs)

    def get_or_post(self, key, default=None):
        for d in (self.request.GET, self.request.POST):
            try:
                return d[key]
            except KeyError:
                pass
        return default

    def params(self, *keys):
        return dict(self.get_param(key) for key in keys)

    def get_param(self, key, type=None):
        semipredicate = self._semipredicate
        if isinstance(key, (list, tuple)):
            key, type = key
            type = self.typemap.get(type, type)
        value = self.get_or_post(key, semipredicate)
        if type and value != semipredicate:
            return key, type(value)
        return key, None if value == semipredicate else value


def simple_get(fun):
    return type(fun.__name__, (ApiView, ), {"get": fun,
                    "__module__": fun.__module__, "__doc__": fun.__doc__})



# Django settings for testproj project.

import os
import sys
# import source code dir
sys.path.insert(0, os.getcwd())
sys.path.insert(0, os.path.join(os.getcwd(), os.pardir))

SITE_ID = 300

DEBUG = True
TEMPLATE_DEBUG = DEBUG

ROOT_URLCONF = "tests.urls"

ADMINS = (
    # ('Your Name', 'your_email@domain.com'),
)

TEST_RUNNER = "django_nose.run_tests"
here = os.path.abspath(os.path.dirname(__file__))
COVERAGE_EXCLUDE_MODULES = ("cyme.tests.*", )

NOSE_ARGS = [os.path.join(here, os.pardir, "cyme", "tests"),
            os.environ.get("NOSE_VERBOSE") and "--verbose" or "",
            "--cover3-package=cyme",
            "--cover3-branch",
            "--cover3-exclude=%s" % ",".join(COVERAGE_EXCLUDE_MODULES)]

BROKER_HOST = "localhost"
BROKER_PORT = 5672
BROKER_VHOST = "/"
BROKER_USER = "guest"
BROKER_PASSWORD = "guest"

TT_HOST = "localhost"
TT_PORT = 1978

CELERY_DEFAULT_EXCHANGE = "test.cyme"
CELERY_DEFAULT_ROUTING_KEY = "test.cyme"
CELERY_DEFAULT_QUEUE = "test.cyme"

CELERY_QUEUES = {"test.cyme": {"binding_key": "test.cyme"}}

MANAGERS = ADMINS

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "cyme-test.db",
    },
}

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django_nose',
    'cyme',
    'cyme.api',
)

CELERY_SEND_TASK_ERROR_EMAILS = False



"""

.. program:: cyme-branch

``cyme-branch``
===============

Starts the cyme branch service.

Options
-------

.. cmdoption:: -i, --id

    Set branch id, if not provided one will be automatically generated.

.. cmdoption:: --without-httpd

    Disable the HTTP server thread.

.. cmdoption:: -l, --loglevel

    Set custom log level. One of DEBUG/INFO/WARNING/ERROR/CRITICAL.
    Default is INFO.

.. cmdoption:: -f, --logfile

    Set custom log file path. Default is :file:`<stderr>`

.. cmdoption:: -D, --instance-dir

    Custom instance directory (default is :file:`instances/``)
    Must be writeable by the user cyme-branch runs as.

.. cmdoption:: -C, --numc

    Number of controllers to start, to handle simultaneous
    requests.  Each controller requires one AMQP connection.
    Default is 2.

.. cmdoption:: --sup-interval

    Supervisor schedule Interval in seconds.  Default is 5.

"""

from __future__ import absolute_import
from __future__ import with_statement

import atexit
import os

from importlib import import_module

from celery import current_app as celery
from celery.bin.base import daemon_options
from celery.platforms import (create_pidlock, detached,
                              signals, set_process_title)
from celery.utils import instantiate
from cl.utils import cached_property, shortuuid

from .base import CymeCommand, Option

BANNER = """
 -------------- cyme@%(id)s v%(version)s
---- **** -----
--- * ***  * -- [Configuration]
-- * - **** ---   . url:         http://%(addr)s:%(port)s
- ** ----------   . broker:      %(broker)s
- ** ----------   . logfile:     %(logfile)s@%(loglevel)s
- ** ----------   . sup:         interval=%(sup.interval)s
- ** ----------   . presence:    interval=%(presence.interval)s
- *** --- * ---   . controllers: #%(controllers)s
-- ******* ----   . instancedir: %(instance_dir)s
--- ***** -----
 -------------- http://celeryproject.org
"""


class Command(CymeCommand):
    branch_cls = "cyme.branch.Branch"
    default_detach_logfile = "b ranch.log"
    default_detach_pidfile = "branch.pid"
    name = "cyme-branch"
    args = '[optional port number, or ipaddr:port]'
    help = 'Starts a cyme branch'
    option_list = tuple(CymeCommand().option_list) + (
        Option('--broker', '-b',
            default=None, action="store", dest="broker",
            help="""Broker URL to use for the cyme message bus.\
                    Default is amqp://guest:guest@localhost:5672//"""),
        Option('--detach',
            default=False, action="store_true", dest="detach",
            help="Detach and run in the background."),
        Option("-i", "--id",
               default=None, action="store", dest="id",
               help="Set explicit branch id."),
        Option("-X", "--no-interaction",
               default=False, action="store_true", dest="no_interaction",
               help="Don't ask questions"),
        Option("--without-httpd",
               default=False, action="store_true", dest="without_httpd",
               help="Disable HTTP server"),
       Option('-l', '--loglevel',
              default="WARNING", action="store", dest="loglevel",
              help="Choose between DEBUG/INFO/WARNING/ERROR/CRITICAL"),
       Option('-D', '--instance-dir',
              default=None, action="store", dest="instance_dir",
              help="Custom instance dir. Default is instances/"),
       Option('-C', '--numc',
              default=2, action="store", type="int", dest="numc",
              help="Number of controllers to start.  Default is 2"),
       Option('--sup-interval',
              default=60, action="store", type="int", dest="sup_interval",
              help="Supervisor schedule interval.  Default is every minute."),
    ) + daemon_options(default_detach_pidfile)

    _startup_pbar = None
    _shutdown_pbar = None

    def handle(self, *args, **kwargs):
        kwargs = self.prepare_options(**kwargs)
        self.loglevel = kwargs.get("loglevel")
        self.logfile = kwargs.get("logfile")
        self.enter_instance_dir()
        self.env.syncdb(interactive=False)
        self.install_cry_handler()
        self.install_rdb_handler()
        self.colored = celery.log.colored(kwargs.get("logfile"))
        self.branch = instantiate(self.branch_cls, *args,
                                 colored=self.colored, **kwargs)
        self.connect_signals()
        print(str(self.colored.cyan(self.banner())))

        self.detached = kwargs.get("detach", False)
        return (self._detach if self.detached else self._start)(**kwargs)

    def setup_default_env(self, env):
        env.setup_eventlet()
        env.setup_pool_limit()

    def stop(self):
        self.set_process_title("shutdown...")

    def on_branch_ready(self, sender=None, **kwargs):
        if self._startup_pbar:
            self._startup_pbar.finish()
            self._startup_pbar = None
        pid = os.getpid()
        self.set_process_title("ready")
        if not self.detached and \
                not self.branch.is_enabled_for("INFO"):
            print("(%s) branch ready" % (pid, ))
        sender.info("[READY] (%s)" % (pid, ))

    def on_branch_shutdown(self, sender=None, **kwargs):
        if self._shutdown_pbar:
            self._shutdown_pbar.finish()
            self._shutdown_pbar = None

    def _detach(self, logfile=None, pidfile=None, uid=None, gid=None,
            umask=None, working_directory=None, **kwargs):
        print("detaching... [pidfile=%s logfile=%s]" % (pidfile, logfile))
        with detached(logfile, pidfile, uid, gid, umask, working_directory):
            return self._start(pidfile=pidfile)

    def _start(self, pidfile=None, **kwargs):
        self.setup_logging(logfile=self.logfile, loglevel=self.loglevel)
        self.set_process_title("boot")
        self.install_signal_handlers()
        if pidfile:
            pidlock = create_pidlock(pidfile).acquire()
            atexit.register(pidlock.release)
        try:
            return self.branch.start().wait()
        except SystemExit:
            self.branch.stop()

    def banner(self):
        branch = self.branch
        addr, port = branch.addrport
        con = branch.controllers
        try:
            pres_interval = con[0].thread.presence.interval
        except AttributeError:
            pres_interval = "(disabled)"
        sup = branch.supervisor.thread
        return BANNER % {"id": branch.id,
                         "version": self.__version__,
                         "broker": branch.connection.as_uri(),
                         "loglevel": self.LOG_LEVELS[branch.loglevel],
                         "logfile": branch.logfile or "[stderr]",
                         "addr": addr or "localhost",
                         "port": port or 8000,
                         "sup.interval": sup.interval,
                         "presence.interval": pres_interval,
                         "controllers": len(con),
                         "instance_dir": self.instance_dir}

    def install_signal_handlers(self):

        def raise_SystemExit(signum, frame):
            raise SystemExit()

        for signal in ("TERM", "INT"):
            signals[signal] = raise_SystemExit

    def set_process_title(self, info):
        set_process_title("%s#%s" % (self.name, shortuuid(self.branch.id)),
                          "%s (-D %s)" % (info, self.instance_dir))

    def repr_controller_id(self, c):
        return shortuuid(c) + c[-2:]

    def connect_signals(self):
        sigs = self.signals
        sigmap = {sigs.branch_startup_request: (self.setup_startup_progress,
                                                self.setup_shutdown_progress),
                  sigs.branch_ready: (self.on_branch_ready, ),
                  sigs.branch_shutdown_complete: (self.on_branch_shutdown, )}
        for sig, handlers in sigmap.iteritems():
            for handler in handlers:
                sig.connect(handler, sender=self.branch)

    def setup_shutdown_progress(self, sender=None, **kwargs):
        from cyme.utils import LazyProgressBar
        if sender.is_enabled_for("DEBUG"):
            return
        c = self.colored
        sigs = (self.signals.thread_pre_shutdown,
                self.signals.thread_pre_join,
                self.signals.thread_post_join,
                self.signals.thread_post_shutdown)
        estimate = (len(sigs) * ((len(sender.components) + 1) * 2)
                    + sum(c.thread.extra_shutdown_steps
                            for c in sender.components))
        text = c.white("Shutdown...").embed()
        p = self._shutdown_pbar = LazyProgressBar(estimate, text,
                                                  c.reset().embed())
        [sig.connect(p.step) for sig in sigs]

    def setup_startup_progress(self, sender=None, **kwargs):
        from cyme.utils import LazyProgressBar
        if sender.is_enabled_for("INFO"):
            return
        c = self.colored
        tsigs = (self.signals.thread_pre_start,
                 self.signals.thread_post_start)
        osigs = (self.signals.httpd_ready,
                 self.signals.supervisor_ready,
                 self.signals.controller_ready,
                 self.signals.branch_ready)

        estimate = (len(tsigs) + ((len(sender.components) + 10) * 2)
                     + len(osigs))
        text = c.white("Startup...").embed()
        p = self._startup_pbar = LazyProgressBar(estimate, text,
                                                 c.reset().embed())
        [sig.connect(p.step) for sig in tsigs + osigs]

    @cached_property
    def signals(self):
        return import_module("cyme.branch.signals")



"""cyme.client.base"""

from __future__ import absolute_import

import anyjson
import requests

from urllib import quote

from celery.datastructures import AttributeDict
from dictshield.document import Document

from .. import __version__, DEBUG
from ..utils import cached_property


class Path(object):

    def __init__(self, s=None, stack=None):
        self.stack = (stack or []) + (s.stack if isinstance(s, Path) else [s])

    def __str__(self, s="/"):
        return (s + s.join(map(quote, filter(None, self.stack))).strip(s) + s)

    def __div__(self, other):
        return Path(other, self.stack)


class Base(object):

    def serialize(self, obj):
        return anyjson.serialize(obj)

    def deserialize(self, text):
        return anyjson.deserialize(text)

    def __getitem__(self, key):
        return self.get(key)

    def __delitem__(self, key):
        return self.delete(key)

    def keys(self):
        return self.all()

    def __iter__(self):
        return iter(self.keys())

    def maybe_async(self, name, nowait):
        if nowait:
            return Path("!") / name
        return name


class Model(Document):

    def __init__(self, parent, *args, **kwargs):
        self.parent = parent
        super(Model, self).__init__(**self._prepare_kwargs(*args, **kwargs))

    def _prepare_kwargs(self, *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], dict):
            kwargs.update(args[0])
        return kwargs

    def delete(self, nowait=False):
        return self.parent.delete(self.name, nowait=nowait)


class Section(Base):
    Model = Model
    name = None
    path = None
    proxy = ["GET", "POST", "PUT", "DELETE"]

    def __init__(self, client):
        self.client = client
        if self.name is None:
            self.name = self.__class__.__name__.lower()
        for attr in self.proxy:
            setattr(self, attr, getattr(self.client, attr))
        self.path = Path(self.name) if self.path is None else self.path

    def all_names(self):
        return self.GET(self.path)

    def all(self):
        return (self.get(name) for name in self.all_names())

    def get(self, name):
        return self.GET(self.path / name, type=self.create_model)

    def add(self, name, nowait=False, **data):
        if isinstance(name, self.Model):
            name = name.name
        return self.POST(self.maybe_async(name, nowait),
                         type=self.create_model, data=data)

    def delete(self, name, nowait=False):
        if isinstance(name, self.Model):
            name = name.name
        return self.DELETE(self.maybe_async(name, nowait))

    def create_model(self, *args, **kwargs):
        model = self.Model(self,
                            **self.Model(self, *args, **kwargs).to_python())
        model.validate()
        return model

    def maybe_async(self, name, nowait):
        if nowait:
            return self.path / "!" / name
        return self.path / name

    def __repr__(self):
        return repr(list(self.all()))


class Client(Base):
    default_url = "http://127.0.0.1:8000"

    def __init__(self, url=None):
        self.url = url.rstrip("/") if url else self.default_url

    def GET(self, path, params=None, type=None):
        return self.request("GET", path, params, None, type)

    def PUT(self, path, params=None, data=None, type=None):
        return self.request("PUT", path, params, data, type)

    def POST(self, path, params=None, data=None, type=None):
        return self.request("POST", path, params, data, type)

    def DELETE(self, path, params=None, data=None, type=None):
        return self.request("DELETE", path, params, data, type)

    def request(self, method, path, params=None, data=None, type=None):
        return self._request(method, self.build_url(path), params, data, type)

    def _prepare(self, d):
        if d:
            return dict((key, value if value is not None else "")
                            for key, value in d.iteritems())

    def _request(self, method, url, params=None, data=None, type=None):
        data = self._prepare(data)
        params = self._prepare(params)
        if DEBUG:
            print("<REQ> %s %r data=%r params=%r" % (method, url,  # noqa+
                                                     data, params))
        type = type or AttributeDict
        r = requests.request(method, str(url),
                             headers=self.headers,
                             params=params, data=data)
        data = None
        if DEBUG:
            data = r.read()
            print("<RES> %r" % (data, ))  # noqa+
        if r.ok:
            if data is None:
                data = r.read()
            ret = self.deserialize(data)
            if isinstance(ret, dict):
                return type(ret)
            return ret
        r.raise_for_status()

    def root(self, method, path=None, params=None, data=None):
        return self._request(method,
                             self.url + str(Path(path) if path else ""),
                             params, data)

    def __repr__(self):
        return "<Client: %r>" % (self.url, )

    @cached_property
    def headers(self):
        return {"Accept": "application/json",
                "User-Agent": "cyme-client:py %r" % (__version__, )}



"""cyme.models"""

from __future__ import absolute_import
from __future__ import with_statement

import errno
import os
import shlex
import warnings

from threading import Lock

from anyjson import deserialize
from celery import current_app as celery
from celery import platforms
from celery.bin.celeryd_multi import MultiTool
from celery.utils.encoding import safe_str
from eventlet import Timeout
from kombu.log import anon_logger
from kombu.pools import connections, producers

from django.db import models
from django.utils.translation import ugettext_lazy as _

from . import managers
from ..utils import cached_property, find_symbol

logger = anon_logger("Instance")


def shsplit(s):
    if s:
        return shlex.split(safe_str(s))
    return []


class Broker(models.Model):
    """Broker connection arguments."""
    # XXX I think this model can be removed now that it only contains the URL.
    objects = managers.BrokerManager()

    url = models.CharField(_(u"URL"), max_length=200, unique=True)

    _pool = None

    class Meta:
        verbose_name = _(u"broker")
        verbose_name_plural = _(u"brokers")

    def __unicode__(self):
        return unicode(self.connection.as_uri())

    def as_dict(self):
        """Returns this broker as a dictionary that can be Json encoded."""
        return {"url": self.url}

    @property
    def pool(self):
        """Connection pool for this connection."""
        return connections[self.connection]

    @property
    def producers(self):
        """Producer pool for this connection."""
        return producers[self.connection]

    @cached_property
    def connection(self):
        return celery.broker_connection(self.url)


class App(models.Model):
    """Application"""
    Broker = Broker
    objects = managers.AppManager()

    name = models.CharField(_(u"name"), max_length=128, unique=True)
    broker = models.ForeignKey(Broker, null=True, blank=True)
    arguments = models.TextField(_(u"arguments"), null=True, blank=True)
    extra_config = models.TextField(_(u"extra config"), null=True, blank=True)

    class Meta:
        verbose_name = _(u"app")
        verbose_name_plural = _(u"apps")

    def __unicode__(self):
        return self.name

    def get_broker(self):
        if self.broker is None:
            return self.Broker._default_manager.get_default()
        return self.broker

    def as_dict(self):
        return {"name": self.name,
                "broker": self.get_broker().url,
                "arguments": self.arguments,
                "extra_config": self.extra_config}


class Queue(models.Model):
    """An AMQP queue that can be consumed from by one or more instances."""
    objects = managers.QueueManager()

    name = models.CharField(_(u"name"), max_length=128, unique=True)
    exchange = models.CharField(_(u"exchange"), max_length=128,
                                default=None, null=True, blank=True)
    exchange_type = models.CharField(_(u"exchange type"), max_length=128,
                                     default=None, null=True, blank=True)
    routing_key = models.CharField(_(u"routing key"), max_length=128,
                                   default=None, null=True, blank=True)
    options = models.TextField(null=True, blank=True)
    is_enabled = models.BooleanField(_(u"is enabled"), default=True)
    created_at = models.DateTimeField(_(u"created at"), auto_now_add=True)

    class Meta:
        verbose_name = _(u"queue")
        verbose_name_plural = _(u"queues")

    def __unicode__(self):
        return self.name

    def as_dict(self):
        """Returns dictionary representation of this queue that can be
        Json encoded."""
        return {"name": self.name,
                "exchange": self.exchange,
                "exchange_type": self.exchange_type,
                "routing_key": self.routing_key,
                "options": self.options}


class Instance(models.Model):
    """A celeryd instance."""
    App = App
    Broker = Broker
    Queue = Queue
    MultiTool = MultiTool

    objects = managers.InstanceManager()
    mutex = Lock()

    app = models.ForeignKey(App)
    name = models.CharField(_(u"name"), max_length=128, unique=True)
    _queues = models.TextField(_(u"queues"), null=True, blank=True)
    max_concurrency = models.IntegerField(_(u"max concurrency"), default=1)
    min_concurrency = models.IntegerField(_(u"min concurrency"), default=1)
    pool = models.CharField(_(u"pool"), max_length=128, blank=True, null=True)
    is_enabled = models.BooleanField(_(u"is enabled"), default=True)
    created_at = models.DateTimeField(_(u"created at"), auto_now_add=True)
    _broker = models.ForeignKey(Broker, null=True, blank=True)
    arguments = models.TextField(_(u"arguments"), null=True, blank=True)
    extra_config = models.TextField(_(u"extra config"), null=True, blank=True)

    class Meta:
        verbose_name = _(u"instance")
        verbose_name_plural = _(u"instances")

    def __unicode__(self):
        return self.name

    def __init__(self, *args, **kwargs):
        app = kwargs.get("app")
        if app is None:
            app = kwargs["app"] = self.App._default_manager.get_default()
        if not isinstance(app, self.App):
            kwargs["app"] = self.App._default_manager.get(name=app)
        super(Instance, self).__init__(*args, **kwargs)

    def as_dict(self):
        """Returns dictionary representation of this instance that
        can be Json encoded."""
        return {"name": self.name,
                "queues": self.queues,
                "max_concurrency": self.max_concurrency,
                "min_concurrency": self.min_concurrency,
                "is_enabled": self.is_enabled,
                "broker": self.broker.url,
                "pool": self.pool,
                "arguments": self.arguments,
                "extra_config": self.extra_config}

    def enable(self):
        """Enables this instance (model-only)."""
        self.is_enabled = True
        self.save()

    def disable(self):
        """Disables this instance (model-only)."""
        self.is_enabled = False
        self.save()

    def start(self, **kwargs):
        """Starts the instance."""
        return self._action("start", **kwargs)

    def stop(self, **kwargs):
        """Shuts down the instance."""
        return self._action("stop", **kwargs)

    def restart(self, **kwargs):
        """Restarts the instance."""
        return self._action("restart", **kwargs)

    def stop_verify(self, **kwargs):
        """Shuts down the instance, waiting until the worker exits"""
        return self._action("stop_verify", **kwargs)

    def alive(self, **kwargs):
        """Returns :const:`True` if the pid responds to signals,
        and the instance responds to ping broadcasts."""
        return self.responds_to_signal() and self.responds_to_ping(**kwargs)

    def stats(self, **kwargs):
        """Returns instance statistics (like ``celeryctl inspect stats``)."""
        return self._query("stats", **kwargs)

    def _update_autoscale(self, max=None, min=None):
        if max is not None:
            self.max_concurrency = max
        if min is not None:
            self.min_concurrency = min
        self.save()
        return [self.max_concurrency, self.min_concurrency]

    def autoscale(self, max=None, min=None, **kwargs):
        """Set max/min autoscale settings."""
        self._update_autoscale(max, min)
        return self._query("autoscale", dict(max=max, min=min), **kwargs)

    def responds_to_ping(self, **kwargs):
        """Returns :const:`True` if the instance responds to
        broadcast ping."""
        return True if self._query("ping", **kwargs) else False

    def responds_to_signal(self):
        """Returns :const:`True` if the pid file exists and the pid
        responds to signals."""
        pid = self.getpid()
        if not pid:
            return False
        try:
            os.kill(pid, 0)
        except OSError, exc:
            if exc.errno == errno.ESRCH:
                return False
            raise
        return True

    def consuming_from(self, **kwargs):
        """Returns the queues the instance is currently consuming from."""
        queues = self._query("active_queues", **kwargs)
        return dict((q["name"], q) for q in queues) if queues else {}

    def add_queue_eventually(self, q):
        """Add queue to instance (model-only)."""
        self.queues.add(q)
        self.save()
        return self

    def remove_queue_eventually(self, q):
        """Remove queue from instance (model-only)."""
        self.queues.remove(q)
        self.save()
        return self

    def add_queue(self, q, **kwargs):
        """Add queue for this instance by name."""
        if isinstance(q, self.Queue):
            q = q.as_dict()
        else:
            queues = find_symbol(self, "..branch.controller.queues")
            try:
                q = queues.get(q)
            except queues.NoRouteError:
                self.queues.remove(q)
                self.save()
                warnings.warn(
                    "Removed unknown consumer: %r from %r" % (q, self.name))
                return
        name = q["name"]
        options = deserialize(q["options"]) if q.get("options") else {}
        exchange = q["exchange"] if q["exchange"] else name
        routing_key = q["routing_key"] if q["routing_key"] else name
        return self._query("add_consumer",
                           dict(queue=q["name"],
                                exchange=exchange,
                                exchange_type=q["exchange_type"],
                                routing_key=routing_key,
                                **options), **kwargs)

    def cancel_queue(self, queue, **kwargs):
        """Cancel queue for this instance by :class:`Queue`."""
        queue = queue.name if isinstance(queue, self.Queue) else queue
        return self._query("cancel_consumer", dict(queue=queue), **kwargs)

    def getpid(self):
        """Get the process id for this instance by reading its pid file.

        Returns :const:`None` if the pid file does not exist.

        """
        return platforms.PIDFile(
                self.pidfile.replace("%n", self.name)).read_pid()

    def get_arguments(self):
        return (list(self.default_args)
              + shsplit(self.app.arguments)
              + shsplit(self.arguments))

    def get_extra_config(self):
        return (list(self.default_config)
              + shsplit(self.app.extra_config)
              + shsplit(self.extra_config))

    def _action(self, action, multi="celeryd-multi"):
        """Execute :program:`celeryd-multi` command."""
        with self.mutex:
            argv = ([multi, action, "--nosplash", '--suffix=""', "--no-color"]
                  + [self.name]
                  + self.get_arguments()
                  + ["--"]
                  + self.get_extra_config())
            logger.info(" ".join(argv))
            return self.multi.execute_from_commandline(argv)

    def _query(self, cmd, args={}, **kwargs):
        """Send remote control command and wait for this instances reply."""
        timeout = kwargs.setdefault("timeout", 3)
        name = self.name
        producer = None
        if "connection" not in kwargs:
            producer = self.broker.producers.acquire(block=True, timeout=3)
            kwargs.update(connection=producer.connection,
                          channel=producer.channel)
        try:
            try:
                with Timeout(timeout):
                    r = celery.control.broadcast(cmd,
                                                 arguments=args, reply=True,
                                                 destination=[name], **kwargs)
            except Timeout:
                return None
            return self.my_reply(r)
        finally:
            if producer is not None:
                producer.release()

    def my_reply(self, replies):
        name = self.name
        for reply in replies or []:
            if name in reply:
                return reply[name]

    @cached_property
    def multi(self):
        env = os.environ.copy()
        env.pop("CELERY_LOADER", None)
        return self.MultiTool(env=env)

    @property
    def direct_queue(self):
        return "dq.%s" % (self.name, )

    @property
    def pidfile(self):
        return self.instance_dir / "worker.pid"

    @property
    def logfile(self):
        return self.instance_dir / "worker.log"

    @property
    def statedb(self):
        return self.instance_dir / "worker.statedb"

    @property
    def default_args(self):
        return ("--broker='%s'" % (self.broker.url, ),
                "--workdir='%s'" % (self.instance_dir, ),
                "--pidfile=%s" % (self.pidfile, ),
                "--logfile='%s'" % (self.logfile, ),
                "--queues='%s'" % (self.direct_queue, ),
                "--statedb='%s'" % (self.statedb, ),
                "--events",
                "--pool='%s'" % (self.pool or self.default_pool, ),
                "--loglevel=INFO",
                "--include=cyme.tasks",
                "--autoscale='%s,%s'" % (self.max_concurrency,
                                       self.min_concurrency))

    @property
    def default_config(self):
        return ("celeryd.prefetch_multiplier=10       \
                 celery.acks_late=yes                 \
                 celery.task_result_expires=3600      \
                 celery.send_task_sent_event=yes".split())

    @property
    def broker(self):
        if self._broker is None:
            return self.app.get_broker()
        return self._broker

    def _get_queues(self):
        instance = self

        class Queues(list):

            def add(self, queue):
                if queue not in self:
                    self.append(queue)
                    self._update_obj()

            def remove(self, queue):
                try:
                    list.remove(self, queue)
                except ValueError:
                    pass
                self._update_obj()

            def as_str(self):

                def mq(q):
                    if isinstance(q, Queue):
                        return q.name
                    return q

                return ",".join(map(mq, self))

            def _update_obj(self):
                instance.queues = self.as_str()

        return Queues(q for q in (self._queues or "").split(",") if q)

    def _set_queues(self, queues):
        if not isinstance(queues, basestring):
            queues = ",".join(queues)
        self._queues = queues

    queues = property(_get_queues, _set_queues)

    @cached_property
    def default_pool(self):
        return find_symbol(self, "..conf.CYME_DEFAULT_POOL")

    @cached_property
    def instance_dir(self):
        dir = find_symbol(self, "..conf.CYME_INSTANCE_DIR") / self.name
        dir.mkdir()
        return dir



"""cyme.branch.intsup

- Internal supervisor used to ensure our threads are still running.

"""

from __future__ import absolute_import, with_statement

from os import _exit
from time import sleep

from cl.g import Event

from .thread import gThread

SUP_ERROR_NOT_STARTED = """\
found thread not able to start?\
"""

SUP_ERROR_PING_TIMEOUT = """\
suspected thread crash or blocking: %r\
"""


class gSup(gThread):

    def __init__(self, thread, signal, interval=5, timeout=600):
        self.thread = thread
        self.interval = interval
        self.timeout = timeout
        self.signal = signal
        super(gSup, self).__init__()

    def start_wait_child(self):
        self._ready_event = Event()
        self.signal.connect(self._on_thread_ready, sender=self.thread)
        self.thread.start()
        self._ready_event.wait()
        assert self._ready_event.ready()
        return self.thread

    def _on_thread_ready(self, **kwargs):
        self._ready_event.send(1)
        self.signal.disconnect(self._on_thread_ready)

    def run(self):
        self.debug("starting")
        interval = self.interval
        thread = self.start_wait_child()
        self.info("started")
        timeout = self.timeout
        critical = self.critical

        while not self.should_stop:
            try:
                pong = thread.ping(timeout)
            except (self.Timeout, Exception), exc:
                critical(SUP_ERROR_PING_TIMEOUT % (exc, ))
                _exit(0)
            if not pong:
                critical(SUP_ERROR_NOT_STARTED)
                _exit(0)
            sleep(interval)

    def stop(self):
        super(gSup, self).stop()
        self.thread.stop()

    @property
    def logger_name(self):
        return "%s<%s>" % (self.name, self.thread.name)



from __future__ import absolute_import

from celery.tests.utils import unittest
from mock import Mock

from cyme.branch import Branch
from cyme.thread import gThread


class test_Branch(unittest.TestCase):

    def setUp(self):
        self._gstart, gThread.start = gThread.start, Mock()

    def tearDown(self):
        gThread.start = self._gstart

    def test_run(self):
        branch = Branch()
        self.assertTrue(branch.id)
        self.assertTrue(branch.components)
        branch2 = Branch(without_httpd=True, without_amqp=True)
        self.assertNotEqual(len(branch2.components), len(branch.components))
        branch.run()
        for component in branch.components:
            component.start.assert_called_with()



"""cyme.bin.cyme"""

from __future__ import absolute_import

from .base import app


@app()
def cyme(env, argv):
    from ..management.commands import cyme
    cyme.Command(env=env).run_from_argv([argv[0], "cyme"] + argv[1:])


if __name__ == "__main__":
    cyme()



"""cyme.management.commands.base"""

from __future__ import absolute_import

import logging
import os
import sys

from optparse import make_option as Option  # noqa

from celery.utils import get_cls_by_name, instantiate
from django.conf import settings
from djcelery.management.base import CeleryCommand
from kombu.log import LOG_LEVELS

from cyme import __version__
from cyme.utils import (cached_property, setup_logging,
                        redirect_stdouts_to_logger)


def die(msg, exitcode=1):
    sys.stderr.write("Error: %s\n" % (msg, ))
    sys.exit(exitcode)


class CymeCommand(CeleryCommand):
    __version__ = __version__
    LOG_LEVELS = LOG_LEVELS

    # see http://code.djangoproject.com/changeset/13319.
    stdout, stderr = sys.stdout, sys.stderr

    default_detach_pidfile = None
    default_detach_logfile = None

    def __init__(self, env=None, *args, **kwargs):
        if env is None:
            env = instantiate("cyme.bin.base.Env")
            self.setup_default_env(env)
        self.env = env

    def setup_default_env(self, env):
        pass

    def get_version(self):
        return "cyme v%s" % (self.__version__, )

    def enter_instance_dir(self):
        self.instance_dir.mkdir(parents=True)
        self.instance_dir.chdir()

    def install_cry_handler(self):
        from celery.apps.worker import install_cry_handler
        return install_cry_handler(logging.getLogger())

    def install_rdb_handler(self):
        from celery.apps.worker import install_rdb_handler
        return install_rdb_handler(envvar="CYME_RDBSIG")

    def redirect_stdouts_to_logger(self, loglevel="INFO", logfile=None,
            redirect_level="WARNING"):
        return redirect_stdouts_to_logger(loglevel, logfile, redirect_level)

    def setup_logging(self, loglevel="WARNING", logfile=None, **kwargs):
        return setup_logging(loglevel, logfile)

    def prepare_options(self, broker=None, loglevel=None, logfile=None,
            pidfile=None, detach=None, instance_dir=None, **kwargs):
        if detach:
            logfile = logfile or self.default_detach_logfile
            pidfile = pidfile or self.default_detach_pidfilE
        if broker:
            settings.BROKER_HOST = broker
        if instance_dir:
            settings.CYME_INSTANCE_DIR = instance_dir
        if pidfile and not os.path.isabs(pidfile):
            pidfile = os.path.join(self.instance_dir, pidfile)
        if logfile and not os.path.isabs(logfile):
            logfile = os.path.join(self.instance_dir, logfile)
        if not isinstance(loglevel, int):
            try:
                loglevel = self.LOG_LEVELS[loglevel.upper()]
            except KeyError:
                self.die("Unknown level %r. Please use one of %s." % (
                        loglevel, "|".join(l for l in self.LOG_LEVELS.keys()
                                    if isinstance(l, basestring))))
        return dict(kwargs, loglevel=loglevel, detach=detach,
                            logfile=logfile, pidfile=pidfile)

    def print_help(self, prog_name=None, subcommand=None):
        return super(CymeCommand, self).print_help("cyme", "")

    @cached_property
    def instance_dir(self):
        return get_cls_by_name("cyme.conf.CYME_INSTANCE_DIR")



"""cyme.branch.httpd

- Our embedded WSGI server used to serve the HTTP API.

"""

from __future__ import absolute_import

from eventlet import listen
from eventlet import wsgi

from django.core.handlers import wsgi as djwsgi
from django.core.servers.basehttp import AdminMediaHandler
from requests import get

from .thread import gThread
from .signals import httpd_ready


class HttpServer(gThread):
    joinable = False

    def __init__(self, addrport=None):
        host, port = addrport or ("", 8000)
        if host == "localhost":
            # dnspython bug?
            host = "127.0.0.1"
        self.host, self.port = self.addrport = (host, port)
        super(HttpServer, self).__init__()

    def server(self, sock, handler):
        return wsgi.server(sock, handler,
                           log=self.create_log(),
                           protocol=self.create_http_protocol())

    def run(self):
        handler = AdminMediaHandler(djwsgi.WSGIHandler())
        sock = listen(self.addrport)
        g = self.spawn(self.server, sock, handler)
        self.info("ready")
        httpd_ready.send(sender=self, addrport=self.addrport,
                         handler=handler, sock=sock)
        return g.wait()

    def _do_ping(self, timeout):
        return get(self.url + "/ping/", timeout=timeout).ok

    def create_log(self):
        logger = self

        class _Log(object):

            def write(self, message):
                message = message.rstrip('\n')
                (logger.debug if '/ping/' in message else logger.info)(message)

        return _Log()

    def create_http_protocol(self):
        logger = self

        class HttpProtocol(wsgi.HttpProtocol):

            def get_format_args(self, format, *args):
                return ["%s - - [%s] %s", self.address_string(),
                                          self.log_date_time_string(),
                                          format] + args

            def log_message(self, format, *args):
                return logger.info(*self.get_format_args(format, *args))

            def log_error(self, format, *args):
                return logger.error(*self.get_format_args(format, *args))

        return HttpProtocol

    @property
    def url(self):
        addr, port = self.addrport
        if not addr or addr in ("0.0.0.0", ):
            addr = "127.0.0.1"
        return "http://%s:%s" % (addr, port)

    @property
    def logger_name(self):
        return "wsgi"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import sys
import codecs

try:
    from setuptools import setup, Command
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, Command  # noqa
from distutils.command.install import INSTALL_SCHEMES

os.environ["CYME_NO_EVAL"] = "yes"
import cyme as distmeta
os.environ.pop("CYME_NO_EVAL", None)
sys.modules.pop("cyme", None)

packages, data_files = [], []
root_dir = os.path.dirname(__file__)
if root_dir != '':
    os.chdir(root_dir)
src_dir = "cyme"


def fullsplit(path, result=None):
    if result is None:
        result = []
    head, tail = os.path.split(path)
    if head == '':
        return [tail] + result
    if head == path:
        return result
    return fullsplit(head, [tail] + result)


for scheme in INSTALL_SCHEMES.values():
    scheme['data'] = scheme['purelib']

SKIP_EXTENSIONS = [".pyc", ".pyo", ".swp", ".swo"]


def is_unwanted_file(filename):
    for skip_ext in SKIP_EXTENSIONS:
        if filename.endswith(skip_ext):
            return True
    return False

for dirpath, dirnames, filenames in os.walk(src_dir):
    # Ignore dirnames that start with '.'
    for i, dirname in enumerate(dirnames):
        if dirname.startswith("."):
            del dirnames[i]
    for filename in filenames:
        if filename.endswith(".py"):
            packages.append('.'.join(fullsplit(dirpath)))
        elif is_unwanted_file(filename):
            pass
        else:
            data_files.append([dirpath, [os.path.join(dirpath, f) for f in
                filenames]])


class RunTests(Command):
    description = "Run the django test suite from the tests dir."
    user_options = []
    extra_env = {}
    extra_args = []

    def run(self):
        for env_name, env_value in self.extra_env.items():
            os.environ[env_name] = str(env_value)

        this_dir = os.getcwd()
        testproj_dir = os.path.join(this_dir, "tests")
        os.chdir(testproj_dir)
        sys.path.append(testproj_dir)
        from django.core.management import execute_manager
        os.environ["DJANGO_SETTINGS_MODULE"] = os.environ.get(
                        "DJANGO_SETTINGS_MODULE", "settings")
        settings_file = os.environ["DJANGO_SETTINGS_MODULE"]
        settings_mod = __import__(settings_file, {}, {}, [''])
        prev_argv = list(sys.argv)
        try:
            sys.argv = [__file__, "test"] + self.extra_args
            execute_manager(settings_mod, argv=sys.argv)
        finally:
            sys.argv = prev_argv

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass


class CIRunTests(RunTests):
    extra_args = ["--with-coverage3", "--with-xunit",
                  "--cover3-xml", "--xunit-file=nosetests.xml",
                  "--cover3-xml-file=coverage.xml"]


if os.path.exists("README.rst"):
    long_description = codecs.open("README.rst", "r", "utf-8").read()
else:
    long_description = "See http://github.com/ask/cyme"


setup(
    name='cyme',
    version=distmeta.__version__,
    description=distmeta.__doc__,
    author=distmeta.__author__,
    author_email=distmeta.__contact__,
    url=distmeta.__homepage__,
    platforms=["any"],
    license="BSD",
    packages=packages,
    data_files=data_files,
    zip_safe=False,
    install_requires=[
        "cl",
        "eventlet",
        "dnspython",
        "Django",
        "django-celery>=2.3.1",
        "requests",
        "dictshield",
        "progressbar"
    ],
    cmdclass={"test": RunTests,
              "citest": CIRunTests},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Framework :: Django",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX",
        "Topic :: Communications",
        "Topic :: System :: Distributed Computing",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    entry_points={
        "console_scripts": [
            "cyme-branch        = cyme.bin.cyme_branch:cyme_branch",
            "cyme               = cyme.bin.cyme:cyme",
            "cyme-list-branches = cyme.bin.cyme_list_branches:main"],
    },
    long_description=long_description,
)



"""cyme.branch.metrics"""

from __future__ import absolute_import

import os
from math import ceil

from ..utils import cached_property


def load_average():
    return tuple(ceil(l * 1e2) / 1e2 for l in os.getloadavg())


class df(object):

    def __init__(self, path):
        self.path = path

    @property
    def total_blocks(self):
        return self.stat.f_blocks * self.stat.f_frsize / 1024

    @property
    def available(self):
        return self.stat.f_bavail * self.stat.f_frsize / 1024

    @property
    def capacity(self):
        avail = self.stat.f_bavail
        used = self.stat.f_blocks - self.stat.f_bfree
        return int(ceil(used * 100.0 / (used + avail) + 0.5))

    @cached_property
    def stat(self):
        return os.statvfs(self.path)
